"""

prepare_classification_script.py

Notebook-y script used to prepare a series of shell commands to run a classifier
(other than MegaClassifier) on a MegaDetector result set.

Differs from prepare_classification_script_mc.py only in the final class mapping step.

"""

#%% Job options

import os

def main():
    organization_name = 'idfg'
    job_name = 'idfg-2022-01-27-EOE2021S_Group6'
    input_filename = 'idfg-2022-01-27-EOE2021S_Group6_detections.filtered_rde_0.60_0.85_30_0.20.json'
    image_base = '/datadrive/idfg/EOE2021S_Group6'
    crop_path = os.path.join(os.path.expanduser('~/crops'),job_name + '_crops')
    device_id = 1

    working_dir_base = os.path.join(os.path.expanduser('~/postprocessing'),
                                                    organization_name,
                                                    job_name)

    output_base = os.path.join(working_dir_base,'combined_api_outputs')

    assert os.path.isdir(working_dir_base)
    assert os.path.isdir(output_base)

    output_file = os.path.join(working_dir_base,'run_idfgclassifier_' + job_name +  '.sh')

    input_files = [
        os.path.join(
            os.path.expanduser('~/postprocessing'),
                            organization_name,
                            job_name,
                            'combined_api_outputs',
                            input_filename
            )
        ]

    for fn in input_files:
        assert os.path.isfile(fn)
        

    #%% Constants

    include_cropping = False

    classifier_base = os.path.expanduser('~/models/camera_traps/idfg_classifier/idfg_classifier_20200905_042558')
    assert os.path.isdir(classifier_base)

    checkpoint_path = os.path.join(classifier_base,'idfg_classifier_ckpt_14_compiled.pt')
    assert os.path.isfile(checkpoint_path)

    classifier_categories_path = os.path.join(classifier_base,'label_index.json')
    assert os.path.isfile(classifier_categories_path)

    classifier_output_suffix = '_idfg_classifier_output.csv.gz'
    final_output_suffix = '_idfgclassifier.json'

    threshold_str = '0.65'
    n_threads_str = '50'
    image_size_str = '300'
    batch_size_str = '64'
    num_workers_str = '8'
    logdir = working_dir_base

    classification_threshold_str = '0.05'

    # This is just passed along to the metadata in the output file, it has no impact
    # on how the classification scripts run.
    typical_classification_threshold_str = '0.75'

    classifier_name = 'idfg4'
            

    #%% Set up environment

    commands = []
    # commands.append('cd MegaDetector/classification\n')
    # commands.append('conda activate cameratraps-classifier\n')


    #%% Crop images

    if include_cropping:
        
        commands.append('\n### Cropping ###\n')
        
        # fn = input_files[0]
        for fn in input_files:
        
            input_file_path = fn
            crop_cmd = ''
            
            crop_comment = '\n# Cropping {}\n'.format(fn)
            crop_cmd += crop_comment
            
            crop_cmd += "python crop_detections.py \\\n" + \
                input_file_path + ' \\\n' + \
                crop_path + ' \\\n' + \
                '--images-dir "' + image_base + '"' + ' \\\n' + \
                '--threshold "' + threshold_str + '"' + ' \\\n' + \
                '--square-crops ' + ' \\\n' + \
                '--threads "' + n_threads_str + '"' + ' \\\n' + \
                '--logdir "' + logdir + '"' + ' \\\n' + \
                '\n'
            crop_cmd = '{}'.format(crop_cmd)
            commands.append(crop_cmd)


    #%% Run classifier

    commands.append('\n### Classifying ###\n')

    # fn = input_files[0]
    for fn in input_files:

        input_file_path = fn
        classifier_output_path = crop_path + classifier_output_suffix
        
        classify_cmd = ''
        
        classify_comment = '\n# Classifying {}\n'.format(fn)
        classify_cmd += classify_comment
        
        classify_cmd += "python run_classifier.py \\\n" + \
            checkpoint_path + ' \\\n' + \
            crop_path + ' \\\n' + \
            classifier_output_path + ' \\\n' + \
            '--detections-json "' + input_file_path + '"' + ' \\\n' + \
            '--classifier-categories "' + classifier_categories_path + '"' + ' \\\n' + \
            '--image-size "' + image_size_str + '"' + ' \\\n' + \
            '--batch-size "' + batch_size_str + '"' + ' \\\n' + \
            '--num-workers "' + num_workers_str + '"' + ' \\\n'
        
        if device_id is not None:
            classify_cmd += '--device {}'.format(device_id)
            
        classify_cmd += '\n\n'    
        classify_cmd = '{}'.format(classify_cmd)
        commands.append(classify_cmd)
            

    #%% Merge classification and detection outputs

    commands.append('\n### Merging ###\n')

    # fn = input_files[0]
    for fn in input_files:

        input_file_path = fn
        classifier_output_path = crop_path + classifier_output_suffix
        final_output_path = os.path.join(output_base,
                                        os.path.basename(classifier_output_path)).\
                                        replace(classifier_output_suffix,
                                        final_output_suffix)
        final_output_path = final_output_path.replace('_detections','')
        final_output_path = final_output_path.replace('_crops','')
        
        merge_cmd = ''
        
        merge_comment = '\n# Merging {}\n'.format(fn)
        merge_cmd += merge_comment
        
        merge_cmd += "python merge_classification_detection_output.py \\\n" + \
            classifier_output_path + ' \\\n' + \
            classifier_categories_path + ' \\\n' + \
            '--output-json "' + final_output_path + '"' + ' \\\n' + \
            '--detection-json "' + input_file_path + '"' + ' \\\n' + \
            '--classifier-name "' + classifier_name + '"' + ' \\\n' + \
            '--threshold "' + classification_threshold_str + '"' + ' \\\n' + \
            '--typical-confidence-threshold "' + typical_classification_threshold_str + '"' + ' \\\n' + \
            '\n'
        merge_cmd = '{}'.format(merge_cmd)
        commands.append(merge_cmd)


    #%% Write everything out

    with open(output_file,'w') as f:
        for s in commands:
            f.write('{}'.format(s))

    import stat
    st = os.stat(output_file)
    os.chmod(output_file, st.st_mode | stat.S_IEXEC)
    
if __name__ == '__main__':
    main()



"""

crop_detections.py

Given a detections JSON file from MegaDetector, crops the bounding boxes above
a certain confidence threshold.

This script takes as input a detections JSON file, usually the output of
detection/run_tf_detector_batch.py or the output of the Batch API in the
"Batch processing API output format".

See https://github.com/agentmorris/MegaDetector/tree/master/api/batch_processing.

The script can crop images that are either available locally or that need to be
downloaded from an Azure Blob Storage container.

We assume that no image contains over 100 bounding boxes, and we always save
crops as RGB .jpg files for consistency. For each image, each bounding box is
cropped and saved to a file with a suffix "___cropXX_mdvY.Y.jpg" added to the
filename as the original image. "XX" ranges from "00" to "99" and "Y.Y"
ndicates the MegaDetector version. Based on the given confidence threshold, we
may skip saving certain bounding box crops, but we still increment the bounding
box number for skipped boxes.

Example cropped image path (with MegaDetector bbox):

  "path/to/image.jpg___crop00_mdv4.1.jpg"

By default, the images are cropped exactly per the given bounding box
coordinates. However, if square crops are desired, pass the --square-crops
flag. This will always generate a square crop whose size is the larger of the
bounding box width or height. In the case that the square crop boundaries exceed
the original image size, the crop is padded with 0s.

This script outputs a log file to:
   
   <output_dir>/crop_detections_log_{timestamp}.json

...which contains images that failed to download and crop properly.

"""

#%% Imports

from __future__ import annotations

import argparse
from collections.abc import Iterable, Mapping, Sequence
from concurrent import futures
from datetime import datetime
import io
import json
import os
from typing import Any, BinaryIO, Optional

from azure.storage.blob import ContainerClient
from PIL import Image, ImageOps
from tqdm import tqdm


#%% Example usage

"""
python crop_detections.py \
    detections.json \
    /path/to/crops \
    --images-dir /path/to/images \
    --container-url "https://account.blob.core.windows.net/container?sastoken" \
    --detector-version "4.1" \
    --threshold 0.8 \
    --save-full-images --square-crops \
    --threads 50 \
    --logdir "."
"""


#%% Main function

def main(detections_json_path: str,
         cropped_images_dir: str,
         images_dir: Optional[str],
         container_url: Optional[str],
         detector_version: Optional[str],
         save_full_images: bool,
         square_crops: bool,
         check_crops_valid: bool,
         confidence_threshold: float,
         threads: int,
         logdir: str) -> None:
    """
    Args:
        detections_json_path: str, path to detections JSON file
        cropped_images_dir: str, path to local directory for saving crops of
            bounding boxes
        images_dir: optional str, path to local directory where images are saved
        container_url: optional str, URL (with SAS token, if necessary) of Azure
            Blob Storage container to download images from, if images are not
            all already locally available in <images_dir>
        detector_version: str, detector version string, e.g., '4.1',
            see {batch_detection_api_url}/supported_model_versions
        save_full_images: bool, whether to save downloaded images to images_dir,
            images_dir must be given if save_full_images=True
        square_crops: bool, whether to crop bounding boxes as squares
        check_crops_valid: bool, whether to load each crop to ensure the file is
            valid (i.e., not truncated)
        confidence_threshold: float, only crop bounding boxes above this value
        threads: int, number of threads to use for downloading images
        logdir: str, path to directory to save log file
    """
    
    # error checking
    assert 0 <= confidence_threshold <= 1, \
            'Invalid confidence threshold {}'.format(confidence_threshold)
    if save_full_images:
        assert images_dir is not None, \
            'save_full_images specified but no images_dir provided'
        if not os.path.exists(images_dir):
            os.makedirs(images_dir, exist_ok=True)
            print(f'Created images_dir at {images_dir}')

    # load detections JSON
    with open(detections_json_path, 'r') as f:
        js = json.load(f)
    detections = {img['file']: img for img in js['images']}
    detection_categories = js['detection_categories']

    # get detector version
    if 'info' in js and 'detector' in js['info']:
        api_det_version = js['info']['detector'] # .rsplit('v', maxsplit=1)[1]
        if detector_version is not None:
            assert api_det_version == detector_version,\
            '.json file specifies a detector version of {}, but the caller has specified {}'.format(
            api_det_version,detector_version)
        else:
            detector_version = api_det_version
    if detector_version is None:
        detector_version = 'unknown'

    # convert from category ID to category name
    images_missing_detections = []

    # copy keys to modify dict in-place
    for img_path in list(detections.keys()):
        info_dict = detections[img_path]
        if 'detections' not in info_dict or info_dict['detections'] is None:
            del detections[img_path]
            images_missing_detections.append(img_path)
            continue
        for d in info_dict['detections']:
            if d['category'] not in detection_categories:
                print('Warning: ignoring detection with category {} for image {}'.format(
                    d['category'],img_path))                
                # This will be removed later when we filter for animals
                d['category'] = 'unsupported'
            else:
                d['category'] = detection_categories[d['category']]

    images_failed_dload_crop, num_downloads, num_crops = download_and_crop(
        detections=detections,
        cropped_images_dir=cropped_images_dir,
        images_dir=images_dir,
        container_url=container_url,
        detector_version=detector_version,
        confidence_threshold=confidence_threshold,
        save_full_images=save_full_images,
        square_crops=square_crops,
        check_crops_valid=check_crops_valid,
        threads=threads)
    print(f'{len(images_failed_dload_crop)} images failed to download or crop.')

    # save log of bad images
    log = {
        'images_missing_detections': images_missing_detections,
        'images_failed_download_or_crop': images_failed_dload_crop,
        'num_new_downloads': num_downloads,
        'num_new_crops': num_crops
    }
    os.makedirs(logdir, exist_ok=True)
    date = datetime.now().strftime('%Y%m%d_%H%M%S')  # e.g., '20200722_110816'
    log_path = os.path.join(logdir, f'crop_detections_log_{date}.json')
    with open(log_path, 'w') as f:
        json.dump(log, f, indent=1)


#%% Support functions

def download_and_crop(
        detections: Mapping[str, Mapping[str, Any]],
        cropped_images_dir: str,
        images_dir: Optional[str],
        container_url: Optional[str],
        detector_version: str,
        confidence_threshold: float,
        save_full_images: bool,
        square_crops: bool,
        check_crops_valid: bool,
        threads: int = 1
        ) -> tuple[list[str], int, int]:
    """
    Saves crops to a file with the same name as the original image with an
    additional suffix appended, starting with 3 underscores:
    - if image has ground truth bboxes: "___cropXX.jpg", where "XX" indicates
        the bounding box index
    - if image has bboxes from MegaDetector: "___cropXX_mdvY.Y.jpg", where
        "Y.Y" indicates the MegaDetector version
    See module docstring for more info and examples.

    Args:
        detections: dict, maps image paths to info dict
            {
                "detections": [{
                    "category": "animal",  # must be name, not "1" or "2"
                    "conf": 0.926,
                    "bbox": [0.0, 0.2762, 0.1539, 0.2825],
                }],
                "is_ground_truth": True  # whether bboxes are ground truth
            }
        cropped_images_dir: str, path to folder where cropped images are saved
        images_dir: optional str, path to folder where full images are saved
        container_url: optional str, URL (with SAS token, if necessary) of Azure
            Blob Storage container to download images from, if images are not
            all already locally available in <images_dir>
        detector_version: str, detector version string, e.g., '4.1'
        confidence_threshold: float, only crop bounding boxes above this value
        save_full_images: bool, whether to save downloaded images to images_dir,
            images_dir must be given and must exist if save_full_images=True
        square_crops: bool, whether to crop bounding boxes as squares
        check_crops_valid: bool, whether to load each crop to ensure the file is
            valid (i.e., not truncated)
        threads: int, number of threads to use for downloading images

    Returns:
        images_failed_download: list of str, images with bounding boxes that
            failed to download or crop properly
        total_downloads: int, number of images downloaded
        total_new_crops: int, number of new crops saved to cropped_images_dir
    """
    
    # True for ground truth, False for MegaDetector
    # always save as .jpg for consistency
    crop_path_template = {
        True: os.path.join(cropped_images_dir, '{img_path}___crop{n:>02d}.jpg'),
        False: os.path.join(
            cropped_images_dir,
            '{img_path}___crop{n:>02d}_' + f'{detector_version}.jpg')
    }

    pool = futures.ThreadPoolExecutor(max_workers=threads)
    future_to_img_path = {}
    images_failed_download = []

    container_client = None
    if container_url is not None:
        container_client = ContainerClient.from_container_url(container_url)

    print(f'Getting bbox info for {len(detections)} images...')
    for img_path in tqdm(sorted(detections.keys())):
        # we already did all error checking above, so we don't do any here
        info_dict = detections[img_path]
        bbox_dicts = info_dict['detections']
        is_ground_truth = info_dict.get('is_ground_truth', False)

        # get the image, either from disk or from Blob Storage
        future = pool.submit(
            load_and_crop, img_path, images_dir, container_client, bbox_dicts,
            confidence_threshold, crop_path_template[is_ground_truth],
            save_full_images, square_crops, check_crops_valid)
        future_to_img_path[future] = img_path

    total = len(future_to_img_path)
    total_downloads = 0
    total_new_crops = 0
    print(f'Reading/downloading {total} images and cropping...')
    for future in tqdm(futures.as_completed(future_to_img_path), total=total):
        img_path = future_to_img_path[future]
        try:
            did_download, num_new_crops = future.result()
            total_downloads += did_download
            total_new_crops += num_new_crops
        except Exception as e:  # pylint: disable=broad-except
            exception_type = type(e).__name__
            tqdm.write(f'{img_path} - generated {exception_type}: {e}')
            images_failed_download.append(img_path)

    pool.shutdown()
    if container_client is not None:
        # inelegant way to close the container_client
        with container_client:
            pass

    print(f'Downloaded {total_downloads} images.')
    print(f'Made {total_new_crops} new crops.')
    return images_failed_download, total_downloads, total_new_crops


def load_local_image(img_path: str |  BinaryIO) -> Optional[Image.Image]:
    """
    Attempts to load an image from a local path.
    """
    
    try:
        with Image.open(img_path) as img:
            img.load()
        return img
    except OSError as e:  # PIL.UnidentifiedImageError is a subclass of OSError
        exception_type = type(e).__name__
        tqdm.write(f'Unable to load {img_path}. {exception_type}: {e}.')
    return None


def load_and_crop(img_path: str,
                  images_dir: Optional[str],
                  container_client: Optional[ContainerClient],
                  bbox_dicts: Iterable[Mapping[str, Any]],
                  confidence_threshold: float,
                  crop_path_template: str,
                  save_full_image: bool,
                  square_crops: bool,
                  check_crops_valid: bool) -> tuple[bool, int]:
    """
    Given an image and a list of bounding boxes, checks if the crops already
    exist. If not, loads the image locally or Azure Blob Storage, then crops it.

    local image path: <images_dir>/<img_path>
    Azure storage: <img_path> as the blob name inside the container

    An image is only downloaded from Azure Blob Storage if it does not already
    exist locally and if it has at least 1 bounding box with confidence greater
    than the confidence threshold.

    Args:
        img_path: str, image path
        images_dir: optional str, path to local directory of images, and where
            full images are saved if save_full_images=True
        container_client: optional ContainerClient, this function does not
            use container_client in any context manager
        bbox_dicts: list of dicts, each dict contains info on a bounding box
        confidence_threshold: float, only crop bounding boxes above this value
        crop_path_template: str, contains placeholders {img_path} and {n}
        save_full_images: bool, whether to save downloaded images to images_dir,
            images_dir must be given and must exist if save_full_images=True
        square_crops: bool, whether to crop bounding boxes as squares
        check_crops_valid: bool, whether to load each crop to ensure the file is
            valid (i.e., not truncated)

    Returns:
        did_download: bool, whether image was downloaded from Azure Blob Storage
        num_new_crops: int, number of new crops successfully saved
    """
    
    did_download = False
    num_new_crops = 0

    # crop_path => normalized bbox coordinates [xmin, ymin, width, height]
    bboxes_tocrop: dict[str, list[float]] = {}
    for i, bbox_dict in enumerate(bbox_dicts):
        # only ground-truth bboxes do not have a "confidence" value
        if 'conf' in bbox_dict and bbox_dict['conf'] < confidence_threshold:
            continue
        if bbox_dict['category'] != 'animal':
            continue
        crop_path = crop_path_template.format(img_path=img_path, n=i)
        if not os.path.exists(crop_path) or (
                check_crops_valid and load_local_image(crop_path) is None):
            bboxes_tocrop[crop_path] = bbox_dict['bbox']
    if len(bboxes_tocrop) == 0:
        return did_download, num_new_crops

    img = None

    # try loading image from local directory
    if images_dir is not None:
        full_img_path = os.path.join(images_dir, img_path)
        debug_path = full_img_path
        if os.path.exists(full_img_path):
            img = load_local_image(full_img_path)

    # try to download image from Blob Storage
    if img is None and container_client is not None:
        debug_path = img_path
        with io.BytesIO() as stream:
            container_client.download_blob(img_path).readinto(stream)
            stream.seek(0)

            if save_full_image:
                os.makedirs(os.path.dirname(full_img_path), exist_ok=True)
                with open(full_img_path, 'wb') as f:
                    f.write(stream.read())
                stream.seek(0)

            img = load_local_image(stream)
        did_download = True

    assert img is not None, 'image "{}" failed to load or download properly'.format(
        debug_path)
    
    if img.mode != 'RGB':
        img = img.convert(mode='RGB')  # always save as RGB for consistency

    # crop the image
    for crop_path, bbox in bboxes_tocrop.items():
        num_new_crops += save_crop(
            img, bbox_norm=bbox, square_crop=square_crops, save=crop_path)
    return did_download, num_new_crops


def save_crop(img: Image.Image, bbox_norm: Sequence[float], square_crop: bool,
              save: str) -> bool:
    """
    Crops an image and saves the crop to file.

    Args:
        img: PIL.Image.Image object, already loaded
        bbox_norm: list or tuple of float, [xmin, ymin, width, height] all in
            normalized coordinates
        square_crop: bool, whether to crop bounding boxes as a square
        save: str, path to save cropped image

    Returns: bool, True if a crop was saved, False otherwise
    """
    
    img_w, img_h = img.size
    xmin = int(bbox_norm[0] * img_w)
    ymin = int(bbox_norm[1] * img_h)
    box_w = int(bbox_norm[2] * img_w)
    box_h = int(bbox_norm[3] * img_h)

    if square_crop:
        # expand box width or height to be square, but limit to img size
        box_size = max(box_w, box_h)
        xmin = max(0, min(
            xmin - int((box_size - box_w) / 2),
            img_w - box_w))
        ymin = max(0, min(
            ymin - int((box_size - box_h) / 2),
            img_h - box_h))
        box_w = min(img_w, box_size)
        box_h = min(img_h, box_size)

    if box_w == 0 or box_h == 0:
        tqdm.write(f'Skipping size-0 crop (w={box_w}, h={box_h}) at {save}')
        return False

    # Image.crop() takes box=[left, upper, right, lower]
    crop = img.crop(box=[xmin, ymin, xmin + box_w, ymin + box_h])

    if square_crop and (box_w != box_h):
        # pad to square using 0s
        crop = ImageOps.pad(crop, size=(box_size, box_size), color=0)

    os.makedirs(os.path.dirname(save), exist_ok=True)
    crop.save(save)
    return True


#%% Command-line driver

def _parse_args() -> argparse.Namespace:
    
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Crop detections from MegaDetector.')
    parser.add_argument(
        'detections_json',
        help='path to detections JSON file')
    parser.add_argument(
        'cropped_images_dir',
        help='path to local directory for saving crops of bounding boxes')
    parser.add_argument(
        '-i', '--images-dir',
        help='path to directory where full images are already available, '
             'or where images will be written if --save-full-images is set')
    parser.add_argument(
        '-c', '--container-url',
        help='URL (including SAS token, if necessary) of Azure Blob Storage '
             'container to download images from, if images are not all already '
             'locally available in <images_dir>')
    parser.add_argument(
        '-v', '--detector-version',
        help='detector version string, e.g., "4.1", used if detector version '
             'cannot be inferred from detections JSON')
    parser.add_argument(
        '--save-full-images', action='store_true',
        help='forces downloading of full images to --images-dir')
    parser.add_argument(
        '--square-crops', action='store_true',
        help='crop bounding boxes as squares')
    parser.add_argument(
        '--check-crops-valid', action='store_true',
        help='load each crop to ensure file is valid (i.e., not truncated)')
    parser.add_argument(
        '-t', '--threshold', type=float, default=0.0,
        help='confidence threshold above which to crop bounding boxes')
    parser.add_argument(
        '-n', '--threads', type=int, default=1,
        help='number of threads to use for downloading and cropping images')
    parser.add_argument(
        '--logdir', default='.',
        help='path to directory to save log file')
    return parser.parse_args()


if __name__ == '__main__':
    
    args = _parse_args()
    main(detections_json_path=args.detections_json,
         cropped_images_dir=args.cropped_images_dir,
         images_dir=args.images_dir,
         container_url=args.container_url,
         detector_version=args.detector_version,
         save_full_images=args.save_full_images,
         square_crops=args.square_crops,
         check_crops_valid=args.check_crops_valid,
         confidence_threshold=args.threshold,
         threads=args.threads,
         logdir=args.logdir)



"""

split_locations_into_train_val.py

Splits a list of location IDs into training and validation, targeting a specific
train/val split for each category, but allowing some categories to be tighter or looser
than others.  Does nothing particularly clever, just randomly splits locations into 
train/val lots of times using the target val fraction, and picks the one that meets the 
specified constraints and minimizes weighted error, where "error" is defined as the
sum of each class's absolute divergence from the target val fraction.

"""

#%% Imports/constants

import random
import numpy as np

from collections import defaultdict
from md_utils.ct_utils import sort_dictionary_by_value
from tqdm import tqdm


#%% Main function

def split_locations_into_train_val(location_to_category_counts,
                                   n_random_seeds=10000,
                                   target_val_fraction=0.15,
                                   category_to_max_allowable_error=None,                                   
                                   category_to_error_weight=None,
                                   default_max_allowable_error=0.1):
    """
    Splits a list of location IDs into training and validation, targeting a specific
    train/val split for each category, but allowing some categories to be tighter or looser
    than others.  Does nothing particularly clever, just randomly splits locations into 
    train/val lots of times using the target val fraction, and picks the one that meets the 
    specified constraints and minimizes weighted error, where "error" is defined as the
    sum of each class's absolute divergence from the target val fraction.    
    
    Args:
        location_to_category_counts (dict): a dict mapping location IDs to dicts,
            with each dict mapping a category name to a count.  Any categories not present 
            in a particular dict are assumed to have a count of zero for that location.
            
            For example:
                
            .. code-block:: none

                {'location-000': {'bear':4,'wolf':10},
                 'location-001': {'bear':12,'elk':20}}
    
        n_random_seeds (int, optional): number of random seeds to try, always starting from zero
        target_val_fraction (float, optional): fraction of images containing each species we'd
            like to put in the val split
        category_to_max_allowable_error (dict, optional): a dict mapping category names
            to maximum allowable errors.  These are hard constraints (i.e., we will error
            if we can't meet them).  Does not need to include all categories; categories not 
            included will be assigned a maximum error according to [default_max_allowable_error].
            If this is None, no hard constraints are applied.
        category_to_error_weight (dict, optional): a dict mapping category names to
            error weights.  You can specify a subset of categories; categories not included here
            have a weight of 1.0.  If None, all categories have the same weight.
        default_max_allowable_error (float, optional): the maximum allowable error for categories not 
            present in [category_to_max_allowable_error].  Set to None (or >= 1.0) to disable hard 
            constraints for categories not present in [category_to_max_allowable_error]
    
    Returns:
        tuple: A two-element tuple:
            - list of location IDs in the val split
            - a dict mapping category names to the fraction of images in the val split                
    """
    
    location_ids = list(location_to_category_counts.keys())
    
    n_val_locations = int(target_val_fraction*len(location_ids))
    
    if category_to_max_allowable_error is None:
        category_to_max_allowable_error = {}
        
    if category_to_error_weight is None:
        category_to_error_weight = {}
        
    # category ID to total count; the total count is used only for printouts
    category_id_to_count = {}
    for location_id in location_to_category_counts:
        for category_id in location_to_category_counts[location_id].keys():
            if category_id not in category_id_to_count:
                category_id_to_count[category_id] = 0
            category_id_to_count[category_id] += \
                location_to_category_counts[location_id][category_id]
    
    category_ids = set(category_id_to_count.keys())
    
    print('Splitting {} categories over {} locations'.format(
        len(category_ids),len(location_ids)))
    
    # random_seed = 0
    def compute_seed_errors(random_seed):
        """
        Computes the per-category error for a specific random seed.
        
        returns weighted_average_error,category_to_val_fraction
        """
        
        # Randomly split into train/val
        random.seed(random_seed)
        val_locations = random.sample(location_ids,k=n_val_locations)
        val_locations_set = set(val_locations)
        
        # For each category, measure the % of images that went into the val set
        category_to_val_fraction = defaultdict(float)
        
        for category_id in category_ids:
            category_val_count = 0
            category_train_count = 0
            for location_id in location_to_category_counts:
                if category_id not in location_to_category_counts[location_id]:
                    location_category_count = 0
                else:
                    location_category_count = location_to_category_counts[location_id][category_id]
                if location_id in val_locations_set:
                    category_val_count += location_category_count
                else:
                    category_train_count += location_category_count
            category_val_fraction = category_val_count / (category_val_count + category_train_count)
            category_to_val_fraction[category_id] = category_val_fraction
        
        # Absolute deviation from the target val fraction for each categorys
        category_errors = {}
        weighted_category_errors = {}
        
        # category = next(iter(category_to_val_fraction))
        for category in category_to_val_fraction:
            
            category_val_fraction = category_to_val_fraction[category]
            
            category_error = abs(category_val_fraction-target_val_fraction)
            category_errors[category] = category_error
        
            category_weight = 1.0
            if category in category_to_error_weight:
                category_weight = category_to_error_weight[category]
            weighted_category_error = category_error * category_weight
            weighted_category_errors[category] = weighted_category_error
        
        weighted_average_error = np.mean(list(weighted_category_errors.values()))
        
        return weighted_average_error,weighted_category_errors,category_to_val_fraction
    
    # ... def compute_seed_errors(...)
    
    # This will only include random seeds that satisfy the hard constraints
    random_seed_to_weighted_average_error = {}
    
    # random_seed = 0
    for random_seed in tqdm(range(0,n_random_seeds)):
        
        weighted_average_error,weighted_category_errors,category_to_val_fraction = \
            compute_seed_errors(random_seed)
            
        seed_satisfies_hard_constraints = True
        
        for category in category_to_val_fraction:
            if category in category_to_max_allowable_error:
                max_allowable_error = category_to_max_allowable_error[category]
            else:
                if default_max_allowable_error is None:
                    continue
                max_allowable_error = default_max_allowable_error
            val_fraction = category_to_val_fraction[category]
            category_error = abs(val_fraction - target_val_fraction)
            if category_error > max_allowable_error:
                seed_satisfies_hard_constraints = False
                break
        
        if seed_satisfies_hard_constraints:            
            random_seed_to_weighted_average_error[random_seed] = weighted_average_error
        
    # ...for each random seed
    
    assert len(random_seed_to_weighted_average_error) > 0, \
        'No random seed met all the hard constraints'
        
    print('\n{} of {} random seeds satisfied hard constraints'.format(
        len(random_seed_to_weighted_average_error),n_random_seeds))
    
    min_error = None
    min_error_seed = None
    
    for random_seed in random_seed_to_weighted_average_error.keys():
        error_metric = random_seed_to_weighted_average_error[random_seed]
        if min_error is None or error_metric < min_error:
            min_error = error_metric
            min_error_seed = random_seed
    
    random.seed(min_error_seed)
    val_locations = random.sample(location_ids,k=n_val_locations)
    train_locations = []
    for location_id in location_ids:
        if location_id not in val_locations:
            train_locations.append(location_id)
            
    print('\nVal locations:\n')        
    for loc in val_locations:
        print('{}'.format(loc))
    print('')
    
    weighted_average_error,weighted_category_errors,category_to_val_fraction = \
        compute_seed_errors(min_error_seed)
        
    random_seed = min_error_seed
    
    category_to_val_fraction = sort_dictionary_by_value(category_to_val_fraction,reverse=True)
    category_to_val_fraction = sort_dictionary_by_value(category_to_val_fraction,
                                                        sort_values=category_id_to_count,
                                                        reverse=True)
    
    
    print('Val fractions by category:\n')
    
    for category in category_to_val_fraction:
        print('{} ({}) {:.2f}'.format(
            category,category_id_to_count[category],
            category_to_val_fraction[category]))
          
    return val_locations,category_to_val_fraction

# ...def split_locations_into_train_val(...)



"""

 save_the_elephants_survey_B.py

 Convert the .csv file provided for the Save the Elephants Survey B data set to a 
 COCO-camera-traps .json file

"""

#%% Constants and environment

from md_visualization import visualize_db
from data_management.databases import integrity_check_json_db
import pandas as pd
import os
import glob
import json
import uuid
import time
import humanfriendly
from PIL import Image
import numpy as np
import logging
from tqdm import tqdm

from md_utils.path_utils import find_images

input_base = r'z:/ste_2019_08_drop'
# input_base = r'/mnt/blobfuse/wildlifeblobssc/ste_2019_08_drop'
input_metadata_file = os.path.join(input_base,'SURVEY B.xlsx')

output_base = r'f:/save_the_elephants/survey_b'
# output_base = r'/home/gramener/survey_b'
output_json_file = os.path.join(output_base,'ste_survey_b.json')
image_directory = os.path.join(input_base,'SURVEY B with False Triggers')
                        
os.makedirs(output_base,exist_ok=True)
assert(os.path.isdir(image_directory))
assert(os.path.isfile(input_metadata_file))

# Handle all unstructured fields in the source data as extra fields in the annotations
mapped_fields = {'No. of Animals in Photo':'num_animals',
                 'No. of new indiviauls (first sighting of new individual)':'num_new_individuals',
                 'Number Adult Males (first sighting of new individual)':'num_adult_males',
                 'Number Adult Females (first sighting of new individual)':'num_adult_females',
                 'Number Adult Unknown (first sighting of new individual)':'num_adult_unknown',
                 'Number Sub-adult Males (first sighting of new individual)':'num_subadult_males',
                 'Number Sub-adult Females (first sighting of new individual)':'num_subadult_females',
                 'Number Sub-adult Unknown (first sighting of new individual)':'num_subadult_unknown',
                 'Number Juvenile (first sighting of new individual)':'num_juvenile',
                 'Number Newborn (first sighting of new individual)':'num_newborn',
                 'Activity':'activity',
                 'Animal ID':'animal_id',
                 'Specific Notes':'notes'}

# photo_type really should be an image property, but there are a few conflicts
# that forced me to handle it as an annotation proprerty
mapped_fields['Photo Type '] = 'photo_type'

#%% Read source data

input_metadata = pd.read_excel(input_metadata_file, sheet_name='9. CT Image')
input_metadata = input_metadata.iloc[2:]

print('Read {} columns and {} rows from metadata file'.format(len(input_metadata.columns),
      len(input_metadata)))


#%% Map filenames to rows, verify image existence

#%% Map filenames to rows, verify image existence

start_time = time.time()

# Maps relative paths to row indices in input_metadata
filenames_to_rows = {}
filenames_with_multiple_annotations = []
missing_images = []

# Build up a map from filenames to a list of rows, checking image existence as we go
for i_row, fn in tqdm(enumerate(input_metadata['Image Name']), total=len(input_metadata)):
    try:
        # Ignore directories
        if not fn.endswith('.JPG'):
            continue

        if fn in filenames_to_rows:
            filenames_with_multiple_annotations.append(fn)
            filenames_to_rows[fn].append(i_row)
        else:
            filenames_to_rows[fn] = [i_row]
            image_path = os.path.join(image_directory, fn)
            if not os.path.isfile(image_path):
                missing_images.append(image_path)
    except:
        continue

elapsed = time.time() - start_time

print('Finished verifying image existence for {} files in {}, found {} filenames with multiple labels, {} missing images'.format(
      len(filenames_to_rows), humanfriendly.format_timespan(elapsed),
      len(filenames_with_multiple_annotations), len(missing_images)))

#%% Make sure the multiple-annotation cases make sense

if False:

    #%%

    fn = filenames_with_multiple_annotations[1000]
    rows = filenames_to_rows[fn]
    assert(len(rows) > 1)
    for i_row in rows:
        print(input_metadata.iloc[i_row]['Species'])

#%% Check for images that aren't included in the metadata file

# Enumerate all images
image_full_paths = find_images(image_directory, bRecursive=True)

unannotated_images = []

for iImage, image_path in tqdm(enumerate(image_full_paths),total=len(image_full_paths)):
    relative_path = os.path.relpath(image_path,image_directory)
    if relative_path not in filenames_to_rows:
        unannotated_images.append(relative_path)

print('Finished checking {} images to make sure they\'re in the metadata, found {} unannotated images'.format(
        len(image_full_paths),len(unannotated_images)))


#%% Create CCT dictionaries

images = []
annotations = []
categories = []

image_ids_to_images = {}

category_name_to_category = {}

# Force the empty category to be ID 0
empty_category = {}
empty_category['name'] = 'empty'
empty_category['id'] = 0
category_name_to_category['empty'] = empty_category
categories.append(empty_category)
next_category_id = 1

start_time = time.time()
# i_image = 0; image_name = list(filenames_to_rows.keys())[i_image]
for image_name in tqdm(list(filenames_to_rows.keys())):

    # Example filename:
    #        
    # 'Site 1_Oloisukut_1\Oloisukut_A11_UP\Service_2\100EK113\EK001382.JPG'
    # 'Site 1_Oloisukut_1\Oloisukut_A11_UP\Service_2.1\100EK113\EK001382.JPG'
    img_id = image_name.replace('\\','/').replace('\n','').replace('/','_').replace(' ','_')
    
    row_indices = filenames_to_rows[image_name]
    
    # i_row = row_indices[0]
    for i_row in row_indices:
        
        row = input_metadata.iloc[i_row]
        assert(row['Image Name'] == image_name)
        try:
            timestamp = row['Date'].strftime("%d/%m/%Y")
        except:
            timestamp = ""
        # timestamp = row['Date']
        station_label = row['Camera Trap Station Label']
        photo_type = row['Photo Type ']
        if isinstance(photo_type,float):
            photo_type = ''
        photo_type = photo_type.strip().lower()
            
        if img_id in image_ids_to_images:
            
            im = image_ids_to_images[img_id]
            assert im['file_name'] == image_name
            assert im['station_label'] == station_label
            
            # There are a small handful of datetime mismatches across annotations
            # for the same image
            # assert im['datetime'] == timestamp
            if im['datetime'] != timestamp:
                print('Warning: timestamp conflict for image {}: {},{}'.format(
                    image_name,im['datetime'],timestamp))
                
        else:
            
            im = {}
            im['id'] = img_id
            im['file_name'] = image_name
            im['datetime'] = timestamp
            im['station_label'] = station_label
            im['photo_type'] = photo_type
            
            image_ids_to_images[img_id] = im
            images.append(im)
    
        species = row['Species']
        
        if (isinstance(species,float) or \
            (isinstance(species,str) and (len(species) == 0))):
            category_name = 'empty'
        elif species.startswith('?'):
            category_name = 'unknown'
        else:
            category_name = species
        
        # Special cases based on the 'photo type' field
        if 'vehicle' in photo_type:
            category_name = 'vehicle'
        # Various spellings of 'community'
        elif 'comm' in photo_type:
            category_name = 'human'
        elif 'camera' in photo_type or 'researcher' in photo_type:
            category_name = 'human'
        elif 'livestock' in photo_type:
            category_name = 'livestock'
        elif 'blank' in photo_type:
            category_name = 'empty'
        elif 'plant movement' in photo_type:
            category_name = 'empty'
            
        category_name = category_name.strip().lower()
            
        # Have we seen this category before?
        if category_name in category_name_to_category:
            category_id = category_name_to_category[category_name]['id'] 
        else:
            category_id = next_category_id
            category = {}
            category['id'] = category_id
            category['name'] = category_name
            category_name_to_category[category_name] = category
            categories.append(category)
            next_category_id += 1
        
        # Create an annotation
        ann = {}        
        ann['id'] = str(uuid.uuid1())
        ann['image_id'] = im['id']    
        ann['category_id'] = category_id
        
        # fieldname = list(mapped_fields.keys())[0]
        for fieldname in mapped_fields:
            target_field = mapped_fields[fieldname]
            val = row[fieldname]
            if isinstance(val,float) and np.isnan(val):
                val = ''
            else:
                val = str(val).strip()
            ann[target_field] = val
            
        annotations.append(ann)
        
    # ...for each row
                
# ...for each image
    
print('Finished creating CCT dictionaries in {}'.format(
      humanfriendly.format_timespan(elapsed)))
    

#%% Create info struct

info = {}
info['year'] = 2019
info['version'] = 1
info['description'] = 'Save the Elephants Survey B'
info['contributor'] = 'Save the Elephants'


#%% Write output

json_data = {}
json_data['images'] = images
json_data['annotations'] = annotations
json_data['categories'] = categories
json_data['info'] = info
json.dump(json_data, open(output_json_file, 'w'), indent=2)

print('Finished writing .json file with {} images, {} annotations, and {} categories'.format(
        len(images),len(annotations),len(categories)))


#%% Validate output

from data_management.databases import integrity_check_json_db

options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = image_directory
options.bCheckImageSizes = False
options.bCheckImageExistence = False
options.bFindUnusedImages = False
    
sortedCategories, data = integrity_check_json_db.integrity_check_json_db(output_json_file,options)


#%% Preview labels

from md_visualization import visualize_db
from data_management.databases import integrity_check_json_db

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 1000
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = True
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
html_output_file,image_db = visualize_db.visualize_db(db_path=output_json_file,
                                                        output_dir=os.path.join(output_base,'preview'),
                                                        image_base_dir=image_directory,
                                                        options=viz_options)
os.startfile(html_output_file)


#%% Scrap

if False:

    pass
    
    #%% Find unique photo types
    
    annotations = image_db['annotations']
    photo_types = set()
    for ann in tqdm(annotations):
        photo_types.add(ann['photo_type'])


"""

 ena24_to_json_2017.py

 Convert the ENA24 data set to a COCO-camera-traps .json file

"""

#%% Constants and environment

import os
import json
import uuid
import time
import humanfriendly
import numpy as np
from PIL import Image
from tqdm import tqdm
import shutil
import zipfile

base_directory = r'e:\wildlife_data\ena24'
output_file = os.path.join(base_directory,'ena24.json')
image_directory = os.path.join(base_directory,'images')
label_directory = os.path.join(base_directory,'labels')

assert(os.path.isdir(label_directory))
assert(os.path.isdir(image_directory))

# Temporary folders for human and non-human images
human_dir = os.path.join(base_directory, 'human')
non_human_dir = os.path.join(base_directory, 'non-human')

human_zipfile = os.path.join(base_directory, 'ena24_humans.zip')
non_human_zipfile = os.path.join(base_directory, 'ena24.zip')

# Clean existing output folders/zipfiles
if os.path.isdir(human_dir):
    shutil.rmtree(human_dir)
if os.path.isdir(non_human_dir):
    shutil.rmtree(non_human_dir)    

if os.path.isfile(human_zipfile):
    os.remove(human_zipfile)
if os.path.isfile(human_zipfile):
    os.remove(non_human_zipfile)
    
os.makedirs(human_dir,exist_ok=True)
os.makedirs(non_human_dir,exist_ok=True)

labels = ['White_Tailed_Deer', 'Dog', 'Bobcat', 'Red Fox', 'Horse', 
          'Domestic Cat', 'American Black Bear', 'Eastern Cottontail', 'Grey Fox', 'Coyote', 
          'Eastern Fox Squirrel', 'Eastern Gray Squirrel', 'Vehicle', 'Eastern Chipmunk', 'Wild Turkey',
          'Northern Raccoon', 'Striped Skunk', 'Woodchuck', 'Virginia Opossum', 'Human', 
          'Bird', 'American Crow', 'Chicken']


#%% Support functions

def zipdir(path, zipfilename, basepath=None):
    """
    Zip everything in [path] into [zipfilename], with paths in the zipfile relative to [basepath]
    """
    ziph = zipfile.ZipFile(zipfilename, 'w', zipfile.ZIP_STORED)
    
    for root, dirs, files in os.walk(path):
        for file in files:
            src = os.path.join(root, file)
            if basepath is None:
                dst = file
            else:
                dst = os.path.relpath(src,basepath)
            ziph.write(src, dst, zipfile.ZIP_STORED)

    ziph.close()


#%% Read source data

image_list = os.listdir(label_directory)
print('Enumerated {} label files'.format(len(image_list)))


#%% Map filenames to rows, verify image existence

startTime = time.time()

# Build up a map from filenames to a list of rows, checking image existence as we go
for filename in image_list:
    imagePath = os.path.join(image_directory, "{}.jpg".format(filename.split(".")[0]))
    assert(os.path.isfile(imagePath))

elapsed = time.time() - startTime
print('Finished verifying image existence for {} files in {}'.format(
      len(image_list),humanfriendly.format_timespan(elapsed)))


#%% Create CCT dictionaries

# Also gets image sizes, so this takes ~6 minutes
#
# Implicitly checks images for overt corruptness, i.e. by not crashing.

images = []
annotations = []

# Map categories to integer IDs (that's what COCO likes)
nextCategoryID = 0
categoriesToCategoryId = {}
categoriesToCounts = {}

# For each image
#
# Because in practice images are 1:1 with annotations in this data set,
# this is also a loop over annotations.

startTime = time.time()
for filename in tqdm(image_list):
    
    contains_human = False    
    im = {}
    im['id'] = filename.split('.')[0]
    fn = "{}.jpg".format(filename.split('.')[0])
    im['file_name'] = fn
    
    # Check image height and width
    imagePath = os.path.join(image_directory, fn)
    assert(os.path.isfile(imagePath))
    pilImage = Image.open(imagePath)
    width, height = pilImage.size
    im['width'] = width
    im['height'] = height

    images.append(im)
    
    label_path = os.path.join(label_directory, filename)
    file_data = open(label_path, 'r').read()
    row = file_data.split()
    category = labels[int(row[0])-1]

    rows = np.loadtxt(label_path)
    
    # Each row is category, [box coordinates]    
    
    # If there's just one row, loadtxt reads it as a 1d array; make it a 2d array 
    # with one row 
    if len(rows.shape)==1:
        rows = rows.reshape(1,-5)
    
    assert (len(rows.shape)==2 and rows.shape[1] == 5)
    
    categories_this_image = set()

    # Each row is a bounding box
    for row in rows:
        
        i_category = int(row[0])-1        
        category = labels[i_category]
        if category == 'Human':
            contains_human = True
        categories_this_image.add(category)
        
        # Have we seen this category before?
        if category in categoriesToCategoryId:
            categoryID = categoriesToCategoryId[category]
            categoriesToCounts[category] += 1
        else:
            categoryID = nextCategoryID
            categoriesToCategoryId[category] = categoryID
            categoriesToCounts[category] = 0
            nextCategoryID += 1
        
        # Create an annotation
        ann = {}
        
        ann['id'] = str(uuid.uuid1())
        ann['image_id'] = im['id']    
        ann['category_id'] = categoryID
        ann['bbox'] = [row[1]*width, row[2]*height, row[3]*width, row[4]*height]
        annotations.append(ann)
        
    # ...for each bounding box
    
    # This was here for debugging; nearly every instance is Human+Horse, Human+Vehicle,
    # or Human+Dog, but there is one Rabbit+Opossium, and a few Deer+Chicken!
    if False:
        if len(categories_this_image) > 1:
            print('Image {} has multiple categories: '.format(filename),end='')
            for c in categories_this_image:
                print(c, end=',')
            print('')
    
    if contains_human:
          shutil.copy(imagePath, os.path.join(base_directory, human_dir))
    else:
          shutil.copy(imagePath, os.path.join(base_directory, non_human_dir))

# ...for each image
    
# Convert categories to a CCT-style dictionary

categories = []

for category in categoriesToCounts:
    print('Category {}, count {}'.format(category, categoriesToCounts[category]))
    categoryID = categoriesToCategoryId[category]
    cat = {}
    cat['name'] = category
    cat['id'] = categoryID
    categories.append(cat)    
    
elapsed = time.time() - startTime
print('Finished creating CCT dictionaries in {}'.format(
      humanfriendly.format_timespan(elapsed)))
    

#%% Create info struct

info = {}
info['year'] = 2016
info['version'] = 1
info['description'] = ''
info['secondary_contributor'] = 'Converted to COCO .json by Vardhan Duvvuri'
info['contributor'] = 'University of Missouri'


#%% Write output

json_data = {}
json_data['images'] = images
json_data['annotations'] = annotations
json_data['categories'] = categories
json_data['info'] = info
json.dump(json_data, open(output_file, 'w'), indent=2)

print('Finished writing .json file with {} images, {} annotations, and {} categories'.format(
        len(images),len(annotations),len(categories)))


#%% Create ZIP files for human and non human

zipdir(human_dir,human_zipfile)
zipdir(non_human_dir,non_human_zipfile)


#%% Validate output

from data_management.databases import integrity_check_json_db

fn = output_file
options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = image_directory
options.bCheckImageSizes = False
options.bCheckImageExistence = True
options.bFindUnusedImages = True
    
sortedCategories, data = integrity_check_json_db.integrity_check_json_db(fn,options)


#%% Preview labels

from md_visualization import visualize_db
from data_management.databases import integrity_check_json_db

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = None
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = True
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
html_output_file,image_db = visualize_db.visualize_db(db_path=output_file,
                                                        output_dir=os.path.join(base_directory,'preview'),
                                                        image_base_dir=image_directory,
                                                        options=viz_options)
os.startfile(html_output_file)



#
# api_backend.py
#
# Defines the model execution service, which pulls requests (one or more images)
# from the shared Redis queue, and runs them through the TF model.
#

#%% Imports

import os
import json
import time
import redis
import argparse
import PIL

from io import BytesIO

from detection.run_detector import load_detector, convert_to_tf_coords
import config
import md_visualization.visualization_utils as vis_utils 

#%% Initialization

db = redis.StrictRedis(host=config.REDIS_HOST, port=config.REDIS_PORT)
current_directory = os.path.dirname(os.path.realpath(__file__))


#%% Main loop

def detect_process():
     
     while True:
        
        # TODO: convert to a blocking read and eliminate the sleep() statement in this loop
        serialized_entry = db.lpop(config.REDIS_QUEUE_NAME)
        all_detection_results = []
        inference_time_detector = []
        
        if serialized_entry:
            
            entry = json.loads(serialized_entry)
            id = entry['id']
            print('Processing images from request id:', id)
            return_confidence_threshold = entry['return_confidence_threshold']

            try:
                
                temp_direc = os.path.join(config.TEMP_FOLDER,id)
                assert os.path.isdir(temp_direc), 'Could not find temporary folder {}'.format(temp_direc)
                
                for filename in os.listdir(temp_direc):
                    
                    image_path = f'{temp_direc}/{filename}'
                    print('Reading image from {}'.format(image_path))
                    image = open(image_path, 'rb')
                    image = vis_utils.load_image(image)

                    start_time = time.time()
                    result = detector.generate_detections_one_image(image, filename, detection_threshold=config.DEFAULT_CONFIDENCE_THRESHOLD)
                    all_detection_results.append(result)

                    elapsed = time.time() - start_time
                    inference_time_detector.append(elapsed)
                    
            except Exception as e:

                print('Detection error: ' + str(e))
                
                db.set(entry['id'], json.dumps({ 
                    'status': 500,
                    'error': 'Detection error: ' + str(e)
                }))

                continue

            # Filter the detections by the confidence threshold
            #
            # Each result is [ymin, xmin, ymax, xmax, confidence, category]
            #
            # Coordinates are relative, with the origin in the upper-left
            detections = {}  
            
            try:
                
                for result in all_detection_results:
                    
                    image_name = result['file']
                    _detections = result.get('detections', None)
                    detections[image_name] = []

                    if _detections is None:
                        continue

                    for d in _detections:
                        if d['conf'] > return_confidence_threshold:
                            res = convert_to_tf_coords(d['bbox'])
                            res.append(d['conf'])
                            res.append(int(d['category']))
                            detections[image_name].append(res)
    
                db.set(entry['id'], json.dumps({ 
                    'status': 200,
                    'detections': detections,
                    'inference_time_detector': inference_time_detector
                }))
              
            except Exception as e:
                print('Error consolidating the detection boxes: ' + str(e))
                
                db.set(entry['id'], json.dumps({ 
                    'status': 500,
                    'error': 'Error consolidating the detection boxes:' + str(e)
                }))

        # ...if serialized_entry
        
        else:
            time.sleep(0.005)
    
    # ...while(True)
    
# ...def detect_process()    
        

#%% Command-line driver

if __name__ == '__main__':
    
    parser = argparse.ArgumentParser(description='api backend')

    # use --non-docker if you are testing without Docker
    #
    # python api_frontend.py --non-docker
    parser.add_argument('--non-docker', action='store_true', default=False)
    args = parser.parse_args()

    if args.non_docker:
        model_path = config.DETECTOR_MODEL_PATH_DEBUG
    else:
        model_path = config.DETECTOR_MODEL_PATH

    detector = load_detector(model_path)

    # run detections on a test image to load the model
    print('Running initial detection to load model...')
    test_image = PIL.Image.new(mode="RGB", size=(200, 200))
    result = detector.generate_detections_one_image(test_image, "test_image", detection_threshold=config.DEFAULT_CONFIDENCE_THRESHOLD)
    print(result)
    print('\n')

    detect_process() 



########
#
# url_utils.py
#
# Frequently-used functions for downloading or manipulating URLs
#
########

#%% Imports and constants

import os
import re
import urllib
import tempfile
import requests

from tqdm import tqdm
from urllib.parse import urlparse

url_utils_temp_dir = None
max_path_len = 255


#%% Download functions

class DownloadProgressBar():
    """
    https://stackoverflow.com/questions/37748105/how-to-use-progressbar-module-with-urlretrieve
    """
    
    def __init__(self):
        self.pbar = None

    def __call__(self, block_num, block_size, total_size):
        if not self.pbar:
            # This is a pretty random import I'd rather not depend on outside of the
            # rare case where it's used, so importing locally
            # pip install progressbar2
            import progressbar            
            self.pbar = progressbar.ProgressBar(max_value=total_size)
            self.pbar.start()
            
        downloaded = block_num * block_size
        if downloaded < total_size:
            self.pbar.update(downloaded)
        else:
            self.pbar.finish()
            

def get_temp_folder(preferred_name='url_utils'):

    global url_utils_temp_dir
    
    if url_utils_temp_dir is None:
        url_utils_temp_dir = os.path.join(tempfile.gettempdir(),preferred_name)
        os.makedirs(url_utils_temp_dir,exist_ok=True)
        
    return url_utils_temp_dir
    
           
def download_url(url, destination_filename=None, progress_updater=None, 
                 force_download=False, verbose=True):
    """
    Download a URL to a file.  If no file is specified, creates a temporary file, 
    with a semi-best-effort to avoid filename collisions.
    
    Prints some diagnostic information and makes sure to omit SAS tokens from printouts.
    
    progress_updater can be "None", "True", or a specific callback.
    """
    
    if progress_updater is not None and isinstance(progress_updater,bool):
        if not progress_updater:
            progress_updater = None
        else:
            progress_updater = DownloadProgressBar()
            
    url_no_sas = url.split('?')[0]
        
    if destination_filename is None:
        target_folder = get_temp_folder()
        url_without_sas = url.split('?', 1)[0]
        
        # This does not guarantee uniqueness, hence "semi-best-effort"
        url_as_filename = re.sub(r'\W+', '', url_without_sas)
        n_folder_chars = len(url_utils_temp_dir)
        if len(url_as_filename) + n_folder_chars > max_path_len:
            print('Warning: truncating filename target to {} characters'.format(max_path_len))
            url_as_filename = url_as_filename[-1*(max_path_len-n_folder_chars):]
        destination_filename = \
            os.path.join(target_folder,url_as_filename)
        
    if (not force_download) and (os.path.isfile(destination_filename)):
        if verbose:
            print('Bypassing download of already-downloaded file {}'.format(os.path.basename(url_no_sas)))
    else:
        if verbose:
            print('Downloading file {} to {}'.format(os.path.basename(url_no_sas),destination_filename),end='')
        target_dir = os.path.dirname(destination_filename)
        os.makedirs(target_dir,exist_ok=True)
        urllib.request.urlretrieve(url, destination_filename, progress_updater)  
        assert(os.path.isfile(destination_filename))
        nBytes = os.path.getsize(destination_filename)
        if verbose:
            print('...done, {} bytes.'.format(nBytes))
        
    return destination_filename


def download_relative_filename(url, output_base, verbose=False):
    """
    Download a URL to output_base, preserving relative path
    """
    
    p = urlparse(url)
    # remove the leading '/'
    assert p.path.startswith('/'); relative_filename = p.path[1:]
    destination_filename = os.path.join(output_base,relative_filename)
    download_url(url, destination_filename, verbose=verbose)


def test_urls(urls, error_on_failure=True):
    """
    Verify that a list of URLs is available (returns status 200).  By default,
    errors if any URL is unavailable.  If error_on_failure is False, returns
    status codes for each URL.
    
    TODO: trivially parallelizable.
    """
    
    status_codes = []
    
    for url in tqdm(urls):
        
        r = requests.get(url)
        
        if error_on_failure and r.status_code != 200:        
            raise ValueError('Could not access {}: error {}'.format(url,r.status_code))
        status_codes.append(r.status_code)
        
    return status_codes


    


"""

 auckland_doc_test_to_json.py

 Convert Auckland DOC data set to COCO camera traps format.  This was
 for a testing data set where a .csv file was provided with class
 information.

"""

#%% Constants and imports

import json
import os
import uuid
import pandas as pd
import datetime
import ntpath
import re
import numpy as np
from tqdm import tqdm

from md_visualization import visualize_db
from data_management.databases import integrity_check_json_db
from md_utils.path_utils import find_images

input_base_dir = r'e:\auckland-test\2_Testing'

input_metadata_file = r'G:\auckland-doc\Maukahuka - Auckland Island - Cat camera data Master April 2019 - DOC-5924483.xlsx'

# Filenames will be stored in the output .json relative to this base dir
output_base_dir = r'g:\auckland-doc'
output_json_filename = os.path.join(output_base_dir, 'auckland-doc-test.json')

assert os.path.isdir(input_base_dir)
os.makedirs(output_base_dir,exist_ok=True)

output_encoding = 'utf-8'
read_image_sizes = True

info = {}
info['year'] = 2020
info['version'] = '1.0'
info['description'] = 'Auckaland DOC Camera Traps (test)'
info['contributor'] = 'Auckland DOC'
info['date_created'] = str(datetime.date.today())


#%% Enumerate files

print('Enumerating files from {}'.format(input_base_dir))
absolute_image_paths = find_images(input_base_dir, recursive=True)
print('Enumerated {} images'.format(len(absolute_image_paths)))

relative_image_paths = []
for fn in absolute_image_paths:
    relative_image_paths.append(os.path.relpath(fn,input_base_dir).replace('\\','/'))

relative_image_paths_set = set(relative_image_paths)

assert len(relative_image_paths_set) == len(relative_image_paths)


#%% Create unique identifier for each image

# The ground truth doesn't have full paths in it; create unique identifiers for each image 
# based on the camera name and filename.
# 
# We store file identifiers as cameraname_filename.
file_identifier_to_relative_paths = {}
camera_names = set()

# relative_path = relative_image_paths[0]
for relative_path in relative_image_paths:
    
    # Example relative paths
    #
    # Summer_Trial_2019/A1_1_42_SD114_20190210/AucklandIsland_A1_1_42_SD114_20190210_01300001.jpg
    # Winter_Trial_2019/Installation/10_F4/10_F4_tmp_201908210001.JPG
    fn = ntpath.basename(relative_path)    
    
    # Find the camera name
    tokens = relative_path.split('/')
    
    if tokens[1] == 'Installation' or 'Rebait' in tokens[1]:
        camera_name = tokens[2]
    
    else:
        # E..g. "A1_1_42_SD114_20190210" in the above example
        camera_token = tokens[1]
        camera_name = None
        m = re.search('^(.+)_SD',camera_token)
        if m:
            camera_name = m.group(1)
        else:
            # For camera tokens like C1_5_D_190207
            m = re.search('^(.+_.+_.+)',camera_token)
            camera_name = m.group(1)
    
    assert camera_name
    camera_names.add(camera_name)
    
    file_identifier = camera_name + '_' + fn
    if file_identifier not in file_identifier_to_relative_paths:
        file_identifier_to_relative_paths[file_identifier] = [relative_path]
    else:
        file_identifier_to_relative_paths[file_identifier].append(relative_path)
    
print('Found {} unique camera names'.format(len(camera_names)))
      
      
#%% Load input data

input_metadata = pd.read_excel(input_metadata_file)

print('Read {} columns and {} rows from metadata file'.format(len(input_metadata.columns),
      len(input_metadata)))

# The spreadsheet has a space after "Camera"
input_metadata = input_metadata.rename(columns={'Camera ':'Camera'})
    

#%% Assemble dictionaries

image_id_to_image = {}
category_name_to_category = {}
annotations = []

# Force the empty category to be ID 0
empty_category = {}
empty_category['name'] = 'empty'
empty_category['id'] = 0
category_name_to_category['empty'] = empty_category

rows_not_found_in_folder = []
rows_ambiguous = []
rows_no_filename = []
rows_no_annotation = []

image_id_to_rows = {}

next_id = 1

category_names = ['cat','mouse','unknown','human','pig','sealion','penguin','dog','openadjusted']

# array([nan, 'Blackbird', 'Bellbird', 'Tomtit', 'Song thrush', 'Pippit',
#       'Pippet', '?', 'Dunnock', 'Song thursh', 'Kakariki', 'Tui', ' ',
#       'Silvereye', 'NZ Pipit', 'Blackbird and Dunnock', 'Unknown',
#       'Pipit', 'Songthrush'], dtype=object)

def bird_name_to_category_name(bird_name):
    bird_name = bird_name.lower().strip().replace(' ','_').replace('song_thursh','song_thrush')
    bird_name = bird_name.replace('pippet','pipt').replace('pippit','pipit').replace('nz_pipit','pipit')
    if bird_name == '?' or bird_name == '' or bird_name == 'unknown':
        category_name = 'unknown_bird'
    else:
        category_name = bird_name
    return category_name
    
bird_names = input_metadata.Bird_ID.unique()
for bird_name in bird_names:
    if isinstance(bird_name,float):
        continue
    category_name = bird_name_to_category_name(bird_name)
    if category_name not in category_names:
        category_names.append(category_name)
        
for category_name in category_names:
    cat = {}
    cat['name'] = category_name
    cat['id'] = next_id
    next_id = next_id +1
    category_name_to_category[category_name] = cat
    
def create_annotation(image_id,category_name,count):
    assert isinstance(image_id,str)
    assert isinstance(category_name,str)
    assert isinstance(count,int) or isinstance(count,float)
    if isinstance(count,float):
        count = int(count)
    ann = {}    
    ann['id'] = str(uuid.uuid1())
    ann['image_id'] = image_id
    category = category_name_to_category[category_name]
    category_id = category['id']
    ann['category_id'] = category_id
    ann['count'] = count
    return ann
    
# i_row = 0; row = input_metadata.iloc[i_row]
for i_row,row in tqdm(input_metadata.iterrows(),total=len(input_metadata)):

    # E.g.: AucklandIsland_A1_1_42_SD114_20190210_01300009.jpg
    filename = row['File']
    if isinstance(filename,float):
        rows_no_filename.append(i_row)
        continue
    
    camera_name = row['Camera']
    file_identifier = camera_name + '_' + filename
    if not file_identifier in file_identifier_to_relative_paths:
        rows_not_found_in_folder.append(i_row)
        continue        

    relative_paths_this_file_id = file_identifier_to_relative_paths[file_identifier]
    
    if len(relative_paths_this_file_id) == 1:
        relative_path = relative_paths_this_file_id[0]
    else:
        
        # We have multiple files matching this identifier, can we uniquely resolve this
        # to one of those files based on the camera ID?
        matches = [s for s in relative_paths_this_file_id if camera_name in s]
        assert len(matches) > 0
        if len(matches) > 1:
            rows_ambiguous.append(i_row)
            continue
        relative_path = matches[0]

    assert filename.endswith('.jpg') or filename.endswith('.JPG')
    image_id = filename.lower().replace('.jpg','')

    if image_id in image_id_to_rows:
        image_id_to_rows[image_id].append(i_row)
        continue
        
    image_id_to_rows[image_id] = [i_row]
        
    im = {}
    im['id'] = image_id 
    im['file_name'] = relative_path
    im['datetime'] = str(row['Date and time'])
    im['camera'] = row['Camera']
    im['sd_card'] = row['SD_Card']
    im['sd_change'] = row['SD_Change']
    im['comments'] = row['Comments']
    
    image_id_to_image[im['id']] = im
    
    # create_annotation(image_id,category_name,count)
    
    # 'SD_Change', 'Cat', 'Mouse', 'Bird', 'Bird_ID', 'False_trig', 'Unknown',
    # 'Human', 'Collared_cat', 'Cat_ID', 'Pig', 'Sea_lion', 'Open_adjusted',
    # 'Penguin', 'Dog', 'Comments', 'Unnamed: 22']
    
    # Each of these categories is handled a little differently...
    
    annotations_this_image = []
    if (not np.isnan(row['Cat'])):
        assert np.isnan(row['Collared_cat'] )
        annotations_this_image.append(create_annotation(im['id'],'cat',row['Cat']))
    
    if (not np.isnan(row['Collared_cat'])):
        assert np.isnan(row['Cat'] )
        annotations_this_image.append(create_annotation(im['id'],'cat',row['Collared_cat']))
    
    if (not np.isnan(row['Bird'])):
        if isinstance(row['Bird_ID'],str):
            category_name = bird_name_to_category_name(row['Bird_ID'])
        else:
            assert np.isnan(row['Bird_ID'])            
            category_name = 'unknown_bird'
        annotations_this_image.append(create_annotation(im['id'],category_name,row['Bird']))
    
    if (not np.isnan(row['False_trig'])):
        annotations_this_image.append(create_annotation(im['id'],'empty',-1))
    
    # These are straightforward
    for s in ['Mouse','Unknown','Pig','Human','Sea_lion','Penguin','Dog','Open_adjusted']:
        if isinstance(row[s],float) or isinstance(row[s],int):
            if not np.isnan(row[s]):
                 annotations_this_image.append(create_annotation(im['id'],s.lower().replace('_',''),row[s]))
        elif isinstance(row[s],str):
            print('File {}, label {}, value {}'.format(filename,s,row[s]))
        else:
            raise ValueError('Error handling count value {}'.format(row[s]))
        
    if len(annotations_this_image) > 1:
        print('Multiple annotations for filename {}'.format(filename))
       
    if len(annotations_this_image) == 0:
        rows_no_annotation.append(i_row)
        
    annotations.extend(annotations_this_image)

# ...for each image


#%% Summarize errors

print('Of {} rows:\n'.format(len(input_metadata)))

print('{} images not found in folder'.format(len(rows_not_found_in_folder)))
print('{} images ambiguously mapped'.format(len(rows_ambiguous)))
print('{} images no filename'.format(len(rows_no_filename)))
print('{} images no annotation'.format(len(rows_no_annotation)))
print('{} images handled successfully, {} total annotations'.format(len(image_id_to_image),len(annotations)))
    

#%% Write output .json

images = list(image_id_to_image.values())
categories = list(category_name_to_category.values())

data = {}
data['info'] = info
data['images'] = images
data['annotations'] = annotations
data['categories'] = categories

json.dump(data, open(output_json_filename, 'w'), indent=1)
print('Finished writing json to {}'.format(output_json_filename))


#%% Validate .json file

options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = input_base_dir
options.bCheckImageSizes = False
options.bCheckImageExistence = False
options.bFindUnusedImages = False

sortedCategories, data, _ = integrity_check_json_db.integrity_check_json_db(output_json_filename, options)


#%% Preview labels

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 2000
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = False
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
viz_options.classes_to_exclude = ['empty']
html_output_file, image_db = visualize_db.visualize_db(db_path=output_json_filename,
                                                         output_dir=os.path.join(
                                                         output_base_dir, 'preview'),
                                                         image_base_dir=input_base_dir,
                                                         options=viz_options)
os.startfile(html_output_file)


#%% Precision-recall analysis

from api.batch_processing.postprocessing.postprocess_batch_results import PostProcessingOptions
from api.batch_processing.postprocessing.postprocess_batch_results import process_batch_results

api_output_file = r'g:\auckland-doc\auckland-doc_20200801\combined_api_outputs\auckland-doc_202008012020.08.01_reformatMaukahuka_Auckland_Island2_TestingSummer_Trial_2019_detections.filtered_rde_0.60_0.85_5_0.05.json'
postprocessing_output_folder = r'G:\auckland-doc\auckland-doc_20200801\postprocessing'
image_base = r'E:\auckland-test\2_Testing'
ground_truth_json_file = output_json_filename

output_base = os.path.join(postprocessing_output_folder,'pr_analysis')
os.makedirs(output_base,exist_ok=True)

options = PostProcessingOptions()
options.unlabeled_classes.append('openadjusted')
options.image_base_dir = image_base
options.parallelize_rendering = True
options.include_almost_detections = True
options.num_images_to_sample = 2500
options.confidence_threshold = 0.75
options.almost_detection_confidence_threshold = 0.7
options.ground_truth_json_file = ground_truth_json_file
options.allow_missing_images = True
options.ground_truth_filename_replacements = {} 
options.api_output_filename_replacements = {'2020.08.01_reformat\\Maukahuka_Auckland_Island\\2_Testing\\':''}
options.api_output_file = api_output_file
options.output_dir = output_base
ppresults = process_batch_results(options)
os.startfile(ppresults.output_html_file)




"""

run_detector.py

Module to run an animal detection model on images.  The main function in this script also renders 
the predicted bounding boxes on images and saves the resulting images (with bounding boxes).

**This script is not a good way to process lots of images**.  It does not produce a useful
output format, and it does not facilitate checkpointing the results so if it crashes you
would have to start from scratch. **If you want to run a detector on lots of images, you should 
check out run_detector_batch.py**.

That said, this script (run_detector.py) is a good way to test our detector on a handful of images 
and get super-satisfying, graphical results.

If you would like to *not* use the GPU on the machine, set the environment
variable CUDA_VISIBLE_DEVICES to "-1".

This script will only consider detections with > 0.005 confidence at all times.
The threshold you provide is only for rendering the results. If you need to
see lower-confidence detections, you can change DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD.

"""

#%% Constants, imports, environment

import argparse
import os
import statistics
import sys
import time
import warnings

import humanfriendly
from tqdm import tqdm

import md_utils.path_utils as path_utils
import md_visualization.visualization_utils as vis_utils

# ignoring all "PIL cannot read EXIF metainfo for the images" warnings
warnings.filterwarnings('ignore', '(Possibly )?corrupt EXIF data', UserWarning)

# Metadata Warning, tag 256 had too many entries: 42, expected 1
warnings.filterwarnings('ignore', 'Metadata warning', UserWarning)

# Numpy FutureWarnings from tensorflow import
warnings.filterwarnings('ignore', category=FutureWarning)

# Useful hack to force CPU inference
# os.environ['CUDA_VISIBLE_DEVICES'] = '-1'


# An enumeration of failure reasons
FAILURE_INFER = 'Failure inference'
FAILURE_IMAGE_OPEN = 'Failure image access'

# Number of decimal places to round to for confidence and bbox coordinates
CONF_DIGITS = 3
COORD_DIGITS = 4

# Label mapping for MegaDetector
DEFAULT_DETECTOR_LABEL_MAP = {
    '1': 'animal',
    '2': 'person',
    '3': 'vehicle'  # available in megadetector v4+
}

# Should we allow classes that don't look anything like the MegaDetector classes?
#
# By default, we error if we see unfamiliar classes.
#
# TODO: the use of a global variable to manage this was fine when this was really
# experimental, but this is really sloppy now that we actually use this code for
# models other than MegaDetector.
USE_MODEL_NATIVE_CLASSES = False

# Each version of the detector is associated with some "typical" values
# that are included in output files, so that downstream applications can 
# use them as defaults.
DETECTOR_METADATA = {
    'v2.0.0':
        {'megadetector_version':'v2.0.0',
         'typical_detection_threshold':0.8,
         'conservative_detection_threshold':0.3},
    'v3.0.0':
        {'megadetector_version':'v3.0.0',
         'typical_detection_threshold':0.8,
         'conservative_detection_threshold':0.3},
    'v4.1.0':
        {'megadetector_version':'v4.1.0',
         'typical_detection_threshold':0.8,
         'conservative_detection_threshold':0.3},
    'v5a.0.0':
        {'megadetector_version':'v5a.0.0',
         'typical_detection_threshold':0.2,
         'conservative_detection_threshold':0.05},
    'v5b.0.0':
        {'megadetector_version':'v5b.0.0',
         'typical_detection_threshold':0.2,
         'conservative_detection_threshold':0.05}
}

DEFAULT_RENDERING_CONFIDENCE_THRESHOLD = DETECTOR_METADATA['v5b.0.0']['typical_detection_threshold']
DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD = 0.005

DEFAULT_BOX_THICKNESS = 4
DEFAULT_BOX_EXPANSION = 0
DEFAULT_LABEL_FONT_SIZE = 16
DETECTION_FILENAME_INSERT = '_detections'

# The model filenames "MDV5A", "MDV5B", and "MDV4" are special; they will trigger an 
# automatic model download to the system temp folder, or they will use the paths specified in the 
# $MDV4, $MDV5A, or $MDV5B environment variables if they exist.
downloadable_models = {
    'MDV4':'https://github.com/agentmorris/MegaDetector/releases/download/v4.1/md_v4.1.0.pb',
    'MDV5A':'https://github.com/agentmorris/MegaDetector/releases/download/v5.0/md_v5a.0.0.pt',
    'MDV5B':'https://github.com/agentmorris/MegaDetector/releases/download/v5.0/md_v5b.0.0.pt'
}

model_string_to_model_version = {
    'v2':'v2.0.0',
    'v3':'v3.0.0',
    'v4.1':'v4.1.0',
    'v5a.0.0':'v5a.0.0',
    'v5b.0.0':'v5b.0.0',
    'mdv5a':'v5a.0.0',
    'mdv5b':'v5b.0.0',
    'mdv4':'v4.1.0',
    'mdv3':'v3.0.0'
}

# Approximate inference speeds (in images per second) for MDv5 based on 
# benchmarks, only used for reporting very coarse expectations about inference time.
device_token_to_mdv5_inference_speed = {
    '4090':17.6,
    '3090':11.4,
    '3080':9.5,
    '3050':4.2,
    'P2000':2.1,
    # These are written this way because they're MDv4 benchmarks, and MDv5
    # is around 3.5x faster than MDv4.
    'V100':2.79*3.5,
    '2080':2.3*3.5,
    '2060':1.6*3.5
}
    

#%% Utility functions

def convert_to_tf_coords(array):
    """
    Converts a bounding box from [x1, y1, width, height] to [y1, x1, y2, x2].  This 
    is mostly not helpful, this function only exists to maintain backwards compatibility
    in the synchronous API, which possibly zero people in the world are using.
    
    Args:
        array (list): a bounding box in [x,y,w,h] format
        
    Returns:
        list: a bounding box in [y1,x1,y2,x2] format
    """
    
    x1 = array[0]
    y1 = array[1]
    width = array[2]
    height = array[3]
    x2 = x1 + width
    y2 = y1 + height
    
    return [y1, x1, y2, x2]


def get_detector_metadata_from_version_string(detector_version):
    """
    Given a MegaDetector version string (e.g. "v4.1.0"), returns the metadata for
    the model.  Used for writing standard defaults to batch output files.
    
    Args:
        detector_version (str): a detection version string, e.g. "v4.1.0", which you
            can extract from a filename using get_detector_version_from_filename()
    
    Returns:
        dict: metadata for this model, suitable for writing to a MD output file
    """
    
    if detector_version not in DETECTOR_METADATA:
        print('Warning: no metadata for unknown detector version {}'.format(detector_version))
        default_detector_metadata = {
            'megadetector_version':'unknown',
            'typical_detection_threshold':0.5,
            'conservative_detection_threshold':0.25
        }
        return default_detector_metadata
    else:
        return DETECTOR_METADATA[detector_version]


def get_detector_version_from_filename(detector_filename):
    r"""
    Gets the version number component of the detector from the model filename.  
    
    [detector_filename] will almost always end with one of the following:
        
    * megadetector_v2.pb
    * megadetector_v3.pb
    * megadetector_v4.1 (not produed by run_detector_batch.py, only found in output files from the deprecated Azure Batch API)
    * md_v4.1.0.pb
    * md_v5a.0.0.pt
    * md_v5b.0.0.pt
    
    This function identifies the version number as "v2.0.0", "v3.0.0", "v4.1.0", 
    "v4.1.0", "v5a.0.0", and "v5b.0.0", respectively.
    
    Args:
        detector_filename (str): model filename, e.g. c:/x/z/md_v5a.0.0.pt
    
    Returns:
        str: a detector version string, e.g. "v5a.0.0", or "multiple" if I'm confused
    """
    
    fn = os.path.basename(detector_filename).lower()
    matches = []
    for s in model_string_to_model_version.keys():
        if s in fn:
            matches.append(s)
    if len(matches) == 0:
        print('Warning: could not determine MegaDetector version for model file {}'.format(detector_filename))
        return 'unknown'
    elif len(matches) > 1:
        print('Warning: multiple MegaDetector versions for model file {}'.format(detector_filename))
        return 'multiple'
    else:
        return model_string_to_model_version[matches[0]]
    

def estimate_md_images_per_second(model_file, device_name=None):
    r"""
    Estimates how fast MegaDetector will run, based on benchmarks.  Defaults to querying
    the current device.  Returns None if no data is available for the current card/model.
    Estimates only available for a small handful of GPUs.  Uses an absurdly simple lookup
    approach, e.g. if the string "4090" appears in the device name, congratulations,
    you have an RTX 4090.
    
    Args:
        model_file (str): model filename, e.g. c:/x/z/md_v5a.0.0.pt
        device_name (str, optional): device name, e.g. blah-blah-4090-blah-blah
    
    Returns:
        float: the approximate number of images this model version can process on this
        device per second
    """
    
    if device_name is None:
        try:
            import torch
            device_name = torch.cuda.get_device_name()
        except Exception as e:
            print('Error querying device name: {}'.format(e))
            return None
    
    model_file = model_file.lower().strip()
    if model_file in model_string_to_model_version.values():
        model_version = model_file
    else:
        model_version = get_detector_version_from_filename(model_file)
        if model_version not in model_string_to_model_version.values():
            print('Error determining model version for model file {}'.format(model_file))
            return None
    
    mdv5_inference_speed = None
    for device_token in device_token_to_mdv5_inference_speed.keys():
        if device_token in device_name:
            mdv5_inference_speed = device_token_to_mdv5_inference_speed[device_token]
            break
    
    if mdv5_inference_speed is None:
        print('No speed estimate available for {}'.format(device_name))
    
    if 'v5' in model_version:
        return mdv5_inference_speed
    elif 'v2' in model_version or 'v3' in model_version or 'v4' in model_version:
        return mdv5_inference_speed / 3.5
    else:
        print('Could not estimate inference speed for model file {}'.format(model_file))
        return None
    
    
def get_typical_confidence_threshold_from_results(results):
    """
    Given the .json data loaded from a MD results file, returns a typical confidence
    threshold based on the detector version.
    
    Args:
        results (dict): a dict of MD results, as it would be loaded from a MD results .json file
    
    Returns:
        float: a sensible default threshold for this model
    """
    
    if 'detector_metadata' in results['info'] and \
        'typical_detection_threshold' in results['info']['detector_metadata']:
        default_threshold = results['info']['detector_metadata']['typical_detection_threshold']
    elif ('detector' not in results['info']) or (results['info']['detector'] is None):
        print('Warning: detector version not available in results file, using MDv5 defaults')
        detector_metadata = get_detector_metadata_from_version_string('v5a.0.0')
        default_threshold = detector_metadata['typical_detection_threshold']
    else:
        print('Warning: detector metadata not available in results file, inferring from MD version')
        detector_filename = results['info']['detector']
        detector_version = get_detector_version_from_filename(detector_filename)
        detector_metadata = get_detector_metadata_from_version_string(detector_version)
        default_threshold = detector_metadata['typical_detection_threshold']

    return default_threshold    

    
def is_gpu_available(model_file):
    r"""
    Determines whether a GPU is available, importing PyTorch or TF depending on the extension
    of model_file.  Does not actually load model_file, just uses that to determine how to check 
    for GPU availability (PT vs. TF).
    
    Args:
        model_file (str): model filename, e.g. c:/x/z/md_v5a.0.0.pt
    
    Returns:
        bool: whether a GPU is available
    """
    
    if model_file.endswith('.pb'):
        import tensorflow.compat.v1 as tf
        gpu_available = tf.test.is_gpu_available()
        print('TensorFlow version:', tf.__version__)
        print('tf.test.is_gpu_available:', gpu_available)                
        return gpu_available
    elif model_file.endswith('.pt'):
        import torch
        gpu_available = torch.cuda.is_available()
        print('PyTorch reports {} available CUDA devices'.format(torch.cuda.device_count()))
        if not gpu_available:
            try:
                # mps backend only available in torch >= 1.12.0
                if torch.backends.mps.is_built and torch.backends.mps.is_available():
                    gpu_available = True
                    print('PyTorch reports Metal Performance Shaders are available')
            except AttributeError:
                pass
        return gpu_available
    else:
        raise ValueError('Unrecognized model file extension for model {}'.format(model_file))


def load_detector(model_file, force_cpu=False):
    r"""
    Loads a TF or PT detector, depending on the extension of model_file.
    
    Args:
        model_file (str): model filename, e.g. c:/x/z/md_v5a.0.0.pt
    
    Returns:
        object: loaded detector object
    """
    
    # Possibly automatically download the model
    model_file = try_download_known_detector(model_file)
    
    start_time = time.time()
    if model_file.endswith('.pb'):
        from detection.tf_detector import TFDetector
        if force_cpu:
            raise ValueError('force_cpu is not currently supported for TF detectors, ' + \
                             'use CUDA_VISIBLE_DEVICES=-1 instead')
        detector = TFDetector(model_file)
    elif model_file.endswith('.pt'):
        from detection.pytorch_detector import PTDetector
        detector = PTDetector(model_file, force_cpu, USE_MODEL_NATIVE_CLASSES)        
    else:
        raise ValueError('Unrecognized model format: {}'.format(model_file))
    elapsed = time.time() - start_time
    print('Loaded model in {}'.format(humanfriendly.format_timespan(elapsed)))
    
    return detector


#%% Main function

def load_and_run_detector(model_file, 
                          image_file_names,
                          output_dir,
                          render_confidence_threshold=DEFAULT_RENDERING_CONFIDENCE_THRESHOLD,
                          crop_images=False, 
                          box_thickness=DEFAULT_BOX_THICKNESS, 
                          box_expansion=DEFAULT_BOX_EXPANSION,
                          image_size=None,
                          label_font_size=DEFAULT_LABEL_FONT_SIZE                          
                          ):
    r"""
    Loads and runs a detector on target images, and visualizes the results.
    
    Args:
        model_file (str): model filename, e.g. c:/x/z/md_v5a.0.0.pt, or a known model
            string, e.g. "MDV5A"
        image_file_names (list): list of absolute paths to process
        output_dir (str): folder to write visualized images to
        render_confidence_threshold (float, optional): only render boxes for detections
            above this threshold
        crop_images (bool, optional): whether to crop detected objects to individual images
            (default is to render images with boxes, rather than cropping)
        box_thickness (float, optional): thickness in pixels for box rendering
        box_expansion (float, optional): box expansion in pixels
        image_size (tuple, optional): image size to use for inference, only mess with this
            if (a) you're using a model other than MegaDetector or (b) you know what you're
            doing
        label_font_size (float, optional): font size to use for displaying class names
            and confidence values in the rendered images        
    """
    
    if len(image_file_names) == 0:
        print('Warning: no files available')
        return

    # Possibly automatically download the model
    model_file = try_download_known_detector(model_file)

    print('GPU available: {}'.format(is_gpu_available(model_file)))
    
    detector = load_detector(model_file)

    detection_results = []
    time_load = []
    time_infer = []

    # Dictionary mapping output file names to a collision-avoidance count.
    #
    # Since we'll be writing a bunch of files to the same folder, we rename
    # as necessary to avoid collisions.
    output_filename_collision_counts = {}

    def input_file_to_detection_file(fn, crop_index=-1):
        """
        Creates unique file names for output files.

        This function does 3 things:
        1) If the --crop flag is used, then each input image may produce several output
            crops. For example, if foo.jpg has 3 detections, then this function should
            get called 3 times, with crop_index taking on 0, 1, then 2. Each time, this
            function appends crop_index to the filename, resulting in
                foo_crop00_detections.jpg
                foo_crop01_detections.jpg
                foo_crop02_detections.jpg

        2) If the --recursive flag is used, then the same file (base)name may appear
            multiple times. However, we output into a single flat folder. To avoid
            filename collisions, we prepend an integer prefix to duplicate filenames:
                foo_crop00_detections.jpg
                0000_foo_crop00_detections.jpg
                0001_foo_crop00_detections.jpg

        3) Prepends the output directory:
                out_dir/foo_crop00_detections.jpg

        Args:
            fn: str, filename
            crop_index: int, crop number

        Returns: output file path
        """
        
        fn = os.path.basename(fn).lower()
        name, ext = os.path.splitext(fn)
        if crop_index >= 0:
            name += '_crop{:0>2d}'.format(crop_index)
        fn = '{}{}{}'.format(name, DETECTION_FILENAME_INSERT, '.jpg')
        if fn in output_filename_collision_counts:
            n_collisions = output_filename_collision_counts[fn]
            fn = '{:0>4d}'.format(n_collisions) + '_' + fn
            output_filename_collision_counts[fn] += 1
        else:
            output_filename_collision_counts[fn] = 0
        fn = os.path.join(output_dir, fn)
        return fn

    # ...def input_file_to_detection_file()
    
    for im_file in tqdm(image_file_names):

        try:
            start_time = time.time()

            image = vis_utils.load_image(im_file)

            elapsed = time.time() - start_time
            time_load.append(elapsed)

        except Exception as e:
            print('Image {} cannot be loaded. Exception: {}'.format(im_file, e))
            result = {
                'file': im_file,
                'failure': FAILURE_IMAGE_OPEN
            }
            detection_results.append(result)
            continue

        try:
            start_time = time.time()

            result = detector.generate_detections_one_image(image, im_file,
                       detection_threshold=DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD,
                       image_size=image_size)
            detection_results.append(result)

            elapsed = time.time() - start_time
            time_infer.append(elapsed)

        except Exception as e:
            print('An error occurred while running the detector on image {}. Exception: {}'.format(im_file, e))
            continue

        try:
            if crop_images:

                images_cropped = vis_utils.crop_image(result['detections'], image,
                                   confidence_threshold=render_confidence_threshold,
                                   expansion=box_expansion)

                for i_crop, cropped_image in enumerate(images_cropped):
                    output_full_path = input_file_to_detection_file(im_file, i_crop)
                    cropped_image.save(output_full_path)

            else:

                # Image is modified in place
                vis_utils.render_detection_bounding_boxes(result['detections'], image,
                            label_map=DEFAULT_DETECTOR_LABEL_MAP,
                            confidence_threshold=render_confidence_threshold,
                            thickness=box_thickness, expansion=box_expansion,
                            label_font_size=label_font_size)
                output_full_path = input_file_to_detection_file(im_file)
                image.save(output_full_path)

        except Exception as e:
            print('Visualizing results on the image {} failed. Exception: {}'.format(im_file, e))
            continue

    # ...for each image

    ave_time_load = statistics.mean(time_load)
    ave_time_infer = statistics.mean(time_infer)
    if len(time_load) > 1 and len(time_infer) > 1:
        std_dev_time_load = humanfriendly.format_timespan(statistics.stdev(time_load))
        std_dev_time_infer = humanfriendly.format_timespan(statistics.stdev(time_infer))
    else:
        std_dev_time_load = 'not available'
        std_dev_time_infer = 'not available'
    print('On average, for each image,')
    print('- loading took {}, std dev is {}'.format(humanfriendly.format_timespan(ave_time_load),
                                                    std_dev_time_load))
    print('- inference took {}, std dev is {}'.format(humanfriendly.format_timespan(ave_time_infer),
                                                      std_dev_time_infer))

# ...def load_and_run_detector()


def download_model(model_name,force_download=False):
    """
    Downloads one of the known models to local temp space if it hasn't already been downloaded.
    
    Args:
        model_name (str): a known model string, e.g. "MDV5A"
        force_download (bool, optional): whether download the model even if the local target 
            file already exists
    """
    
    import tempfile
    from md_utils.url_utils import download_url
    model_tempdir = os.path.join(tempfile.gettempdir(), 'megadetector_models')    
    os.makedirs(model_tempdir,exist_ok=True)
    
    # This is a lazy fix to an issue... if multiple users run this script, the
    # "megadetector_models" folder is owned by the first person who creates it, and others
    # can't write to it.  I could create uniquely-named folders, but I philosophically prefer
    # to put all the individual UUID-named folders within a larger folder, so as to be a 
    # good tempdir citizen.  So, the lazy fix is to make this world-writable.
    try:
        os.chmod(model_tempdir,0o777)
    except Exception:
        pass
    if model_name not in downloadable_models:
        print('Unrecognized downloadable model {}'.format(model_name))
        return None
    url = downloadable_models[model_name]
    destination_filename = os.path.join(model_tempdir,url.split('/')[-1])
    local_file = download_url(url, destination_filename=destination_filename, progress_updater=None, 
                     force_download=force_download, verbose=True)
    return local_file


def try_download_known_detector(detector_file):
    """
    Checks whether detector_file is really the name of a known model, in which case we will
    either read the actual filename from the corresponding environment variable or download
    (if necessary) to local temp space.  Otherwise just returns the input string.
    
    Args:
        detector_file (str): a known model string (e.g. "MDV5A"), or any other string (in which
            case this function is a no-op)
    
    Returns:
        str: the local filename to which the model was downloaded, or the same string that
        was passed in, if it's not recognized as a well-known model name
    """
    
    if detector_file in downloadable_models:
        if detector_file in os.environ:
            fn = os.environ[detector_file]
            print('Reading MD location from environment variable {}: {}'.format(
                detector_file,fn))
            detector_file = fn
        else:
            print('Downloading model {}'.format(detector_file))
            detector_file = download_model(detector_file)
    return detector_file
        
    

#%% Command-line driver

def main():

    parser = argparse.ArgumentParser(
        description='Module to run an animal detection model on images')
    
    parser.add_argument(
        'detector_file',
        help='Path detector model file (.pb or .pt).  Can also be MDV4, MDV5A, or MDV5B to request automatic download.')
    
    # Must specify either an image file or a directory
    group = parser.add_mutually_exclusive_group(required=True)    
    group.add_argument(
        '--image_file',
        type=str,
        default=None,
        help='Single file to process, mutually exclusive with --image_dir')
    group.add_argument(
        '--image_dir',
        type=str,
        default=None,
        help='Directory to search for images, with optional recursion by adding --recursive')
    
    parser.add_argument(
        '--recursive',
        action='store_true',
        help='Recurse into directories, only meaningful if using --image_dir')
    
    parser.add_argument(
        '--output_dir',
        type=str,
        default=None,
        help='Directory for output images (defaults to same as input)')
    
    parser.add_argument(
        '--image_size',
        type=int,
        default=None,
        help=('Force image resizing to a (square) integer size (not recommended to change this)'))
    
    parser.add_argument(
        '--threshold',
        type=float,
        default=DEFAULT_RENDERING_CONFIDENCE_THRESHOLD,
        help=('Confidence threshold between 0 and 1.0; only render' + 
              ' boxes above this confidence (defaults to {})'.format(
              DEFAULT_RENDERING_CONFIDENCE_THRESHOLD)))
    
    parser.add_argument(
        '--crop',
        default=False,
        action='store_true',
        help=('If set, produces separate output images for each crop, '
              'rather than adding bounding boxes to the original image'))
    
    parser.add_argument(
        '--box_thickness',
        type=int,
        default=DEFAULT_BOX_THICKNESS,
        help=('Line width (in pixels) for box rendering (defaults to {})'.format(
              DEFAULT_BOX_THICKNESS)))
    
    parser.add_argument(
        '--box_expansion',
        type=int,
        default=DEFAULT_BOX_EXPANSION,
        help=('Number of pixels to expand boxes by (defaults to {})'.format(
              DEFAULT_BOX_EXPANSION)))
    
    parser.add_argument(
        '--label_font_size',
        type=int,
        default=DEFAULT_LABEL_FONT_SIZE,
        help=('Label font size (defaults to {})'.format(
              DEFAULT_LABEL_FONT_SIZE)))
    
    parser.add_argument(
        '--process_likely_output_images',
        action='store_true',
        help=('By default, we skip images that end in {}, because they probably came from this script. '\
              .format(DETECTION_FILENAME_INSERT) + \
              'This option disables that behavior.'))
    
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    # If the specified detector file is really the name of a known model, find 
    # (and possibly download) that model
    args.detector_file = try_download_known_detector(args.detector_file)
    
    assert os.path.exists(args.detector_file), 'detector file {} does not exist'.format(
        args.detector_file)
    assert 0.0 < args.threshold <= 1.0, 'Confidence threshold needs to be between 0 and 1'

    if args.image_file:
        image_file_names = [args.image_file]
    else:
        image_file_names = path_utils.find_images(args.image_dir, args.recursive)

    # Optionally skip images that were probably generated by this script
    if not args.process_likely_output_images:
        image_file_names_valid = []
        for fn in image_file_names:
            if os.path.splitext(fn)[0].endswith(DETECTION_FILENAME_INSERT):
                print('Skipping likely output image {}'.format(fn))
            else:
                image_file_names_valid.append(fn)
        image_file_names = image_file_names_valid
                
    print('Running detector on {} images...'.format(len(image_file_names)))

    if args.output_dir:
        os.makedirs(args.output_dir, exist_ok=True)
    else:
        if args.image_dir:
            args.output_dir = args.image_dir
        else:
            # but for a single image, args.image_dir is also None
            args.output_dir = os.path.dirname(args.image_file)

    load_and_run_detector(model_file=args.detector_file,
                          image_file_names=image_file_names,
                          output_dir=args.output_dir,
                          render_confidence_threshold=args.threshold,
                          box_thickness=args.box_thickness,
                          box_expansion=args.box_expansion,                          
                          crop_images=args.crop,
                          image_size=args.image_size,
                          label_font_size=args.label_font_size)

if __name__ == '__main__':
    main()


#%% Interactive driver

if False:

    #%%
    model_file = r'c:\temp\models\md_v4.1.0.pb'
    image_file_names = path_utils.find_images(r'c:\temp\demo_images\ssverymini')
    output_dir = r'c:\temp\demo_images\ssverymini'
    render_confidence_threshold = 0.8
    crop_images = True

    load_and_run_detector(model_file=model_file,
                          image_file_names=image_file_names,
                          output_dir=output_dir,
                          render_confidence_threshold=render_confidence_threshold,
                          crop_images=crop_images)



"""

train_utils.py

Utility functions useful for training a classifier.

This script should NOT depend on any other file within this repo. It should
especially be agnostic to PyTorch vs. TensorFlow.

"""

#%% Imports

from __future__ import annotations

from collections.abc import Mapping, Sequence
import dataclasses
import heapq
import io
import json
from typing import Any, Optional

import matplotlib.figure
import numpy as np
import pandas as pd
import PIL.Image
import scipy.interpolate


#%% Classes

@dataclasses.dataclass(order=True)
class HeapItem:
    """
    A wrapper over non-comparable data with a comparable priority value.
    """
    
    priority: Any
    data: Any = dataclasses.field(compare=False, repr=False)


def add_to_heap(h: list[Any], item: HeapItem, k: Optional[int] = None) -> None:
    """
    Tracks the max k elements using a heap.

    We use a min-heap for this task. When a new element comes in, we compare it
    to the smallest node in the heap, h[0]. If the new value is greater than
    h[0], we pop h[0] and add the new element in.

    Args:
        h: list, either empty [] or already heapified
        item: HeapItem
        k: int, desired capacity of the heap, or None for no limit
    """
    
    if k is None or len(h) < k:
        heapq.heappush(h, item)
    else:
        heapq.heappushpop(h, item)


#%% Functions

def prefix_all_keys(d: Mapping[str, Any], prefix: str) -> dict[str, Any]:
    """
    Returns a new dict where the keys are prefixed by <prefix>.
    """
    
    return {f'{prefix}{k}': v for k, v in d.items()}


def fig_to_img(fig: matplotlib.figure.Figure) -> np.ndarray:
    """
    Converts a matplotlib figure to an image represented by a numpy array.
    
    Returns: np.ndarray, type uint8, shape [H, W, 3]
    """
    
    with io.BytesIO() as b:
        fig.savefig(b, transparent=False, bbox_inches='tight', pad_inches=0,
                    format='png')
        b.seek(0)
        fig_img = np.asarray(PIL.Image.open(b).convert('RGB'))
    assert fig_img.dtype == np.uint8
    return fig_img


def imgs_with_confidences(imgs_list: list[tuple[Any, ...]],
                          label_names: Sequence[str],
                          ) -> tuple[matplotlib.figure.Figure, list[str]]:
    """
    Args:
        imgs_list: list of tuple, each tuple consists of:
            img: array_like, shape [H, W, C], type either float [0, 1] or uint8
            label_id: int, label index
            topk_conf: list of float, confidence scores for topk predictions
            topk_preds: list of int, label indices for topk predictions
            img_file: str, path to image file
        label_names: list of str, label names in order of label id

    Returns:
        fig: matplotlib.figure.Figure
        img_files: list of str
    """
    
    imgs, img_files, tags, titles = [], [], [], []
    for img, label_id, topk_conf, topk_preds, img_file in imgs_list:
        imgs.append(img)
        img_files.append(img_file)
        tags.append(label_names[label_id])

        lines = []
        for pred, conf in zip(topk_preds, topk_conf):
            pred_name = label_names[pred]
            lines.append(f'{pred_name}: {conf:.03f}')
        titles.append('\n'.join(lines))

    fig = plot_img_grid(imgs=imgs, row_h=3, col_w=2.5, tags=tags, titles=titles)
    return fig, img_files


def plot_img_grid(imgs: Sequence[Any], row_h: float, col_w: float,
                  ncols: Optional[int] = None,
                  tags: Optional[Sequence[str]] = None,
                  titles: Optional[Sequence[str]] = None
                  ) -> matplotlib.figure.Figure:
    """
    Plots a grid of images.

    Args:
        imgs: list of images, each image is either an array or a PIL Image,
            see matplotlib.axes.Axes.imshow() documentation for supported shapes
        row_h: float, row height in inches
        col_w: float, col width in inches
        ncols: optional int, number of columns, defaults to len(imgs)
        tags: optional list of str, tags are displayed in upper-left corner of
            each image on a white background
        titles: optional list of str, text displayed above each image

    Returns: matplotlib.figure.Figure
    """
    
    # input validation
    num_images = len(imgs)
    if tags is not None:
        assert len(tags) == len(imgs)
    if titles is not None:
        assert len(titles) == len(imgs)

    if ncols is None:
        ncols = num_images

    nrows = int(np.ceil(len(imgs) / ncols))
    fig = matplotlib.figure.Figure(figsize=(ncols * col_w, nrows * row_h),
                                   tight_layout=True)
    axs = fig.subplots(nrows, ncols, squeeze=False)

    # plot the images
    for i in range(num_images):
        r, c = i // ncols, i % ncols
        ax = axs[r, c]
        ax.imshow(imgs[i])
        if tags is not None:
            ax.text(-0.2, -0.2, tags[i], ha='left', va='top',
                    bbox=dict(lw=0, facecolor='white'))
        if titles is not None:
            ax.set_title(titles[i])

    # adjust the figure
    for r in range(nrows):
        for c in range(ncols):
            axs[r, c].set_axis_off()
            axs[r, c].set_aspect('equal')
    fig.subplots_adjust(wspace=0, hspace=0)
    return fig


def load_splits(splits_json_path: str) -> dict[str, set[tuple[str, str]]]:
    """
    Loads location splits from JSON file and assert that there are no
    overlaps between splits.

    Args:
        splits_json_path: str, path to JSON file

    Returns: dict, maps split to set of (dataset, location) tuples
    """
    
    with open(splits_json_path, 'r') as f:
        split_to_locs_js = json.load(f)
    split_to_locs = {
        split: set((loc[0], loc[1]) for loc in locs)
        for split, locs in split_to_locs_js.items()
    }
    assert split_to_locs['train'].isdisjoint(split_to_locs['val'])
    assert split_to_locs['train'].isdisjoint(split_to_locs['test'])
    assert split_to_locs['val'].isdisjoint(split_to_locs['test'])
    return split_to_locs


def load_dataset_csv(dataset_csv_path: str,
                     label_index_json_path: str,
                     splits_json_path: str,
                     multilabel: bool,
                     weight_by_detection_conf: bool | str,
                     label_weighted: bool
                     ) -> tuple[pd.DataFrame,
                                list[str],
                                dict[str, set[tuple[str, str]]]
                               ]:
    """
    Args:
        dataset_csv_path: str, path to CSV file with columns
            ['dataset', 'location', 'label', 'confidence'], where label is a
            comma-delimited list of labels
        label_index_json_path: str, path to label index JSON file
        splits_json_path: str, path to splits JSON file
        multilabel: bool, whether a single example can have multiple labels
        weight_by_detection_conf: bool or str
            - if True: assumes classification CSV's 'confidence' column
                represents calibrated scores
            - if str: path the .npz file containing x/y values for isotonic
                regression calibration function
        label_weighted: bool, whether to give each label equal weight

    Returns:
        df: pd.DataFrame, with columns
            dataset_location: tuples of (dataset, location)
            label: str if not multilabel, list of str if multilabel
            label_index: int if not multilabel, list of int if multilabel
            weights: float, weight for each example
                column exists if and only if label_weighted=True or
                weight_by_detection_conf is not False
        label_names: list of str, label names in order of label id
        split_to_locs: dict, maps split to set of (dataset, location) tuples
    """
    
    # read in dataset CSV and create merged (dataset, location) col
    df = pd.read_csv(dataset_csv_path, index_col=False, float_precision='high')
    df['dataset_location'] = list(zip(df['dataset'], df['location']))

    with open(label_index_json_path, 'r') as f:
        idx_to_label = json.load(f)
    label_names = [idx_to_label[str(i)] for i in range(len(idx_to_label))]
    label_to_idx = {label: idx for idx, label in enumerate(label_names)}

    # map label to label_index
    if multilabel:
        df['label'] = df['label'].map(lambda x: x.split(','))
        df['label_index'] = df['label'].map(
            lambda labellist: tuple(sorted(label_to_idx[y] for y in labellist)))
    else:
        assert not any(df['label'].str.contains(','))
        df['label_index'] = df['label'].map(label_to_idx.__getitem__)

    # load the splits
    split_to_locs = load_splits(splits_json_path)

    if weight_by_detection_conf:
        df['weights'] = 1.0

        # only weight the training set by detection confidence
        train_mask = df['dataset_location'].isin(split_to_locs['train'])
        df.loc[train_mask, 'weights'] = df.loc[train_mask, 'confidence']

        if isinstance(weight_by_detection_conf, str):
            # isotonic regression calibration of MegaDetector confidence
            with np.load(weight_by_detection_conf) as npz:
                calib = scipy.interpolate.interp1d(
                    x=npz['x'], y=npz['y'], kind='linear')
            df.loc[train_mask, 'weights'] = calib(df.loc[train_mask, 'weights'])

    if label_weighted:
        if multilabel:
            raise NotImplementedError

        if 'weights' not in df.columns:
            df['weights'] = 1.0

        # treat each split separately
        # new_weight[i] = confidence[i] * (n / c) / total_confidence(i's label)
        # - n = # examples in split (weighted by confidence); c = # labels
        # - weight allocated to each label is n/c
        # - within each label, weigh each example proportional to confidence
        # - new weights sum to n
        c = len(label_names)
        for split, locs in split_to_locs.items():
            split_mask = df['dataset_location'].isin(locs)
            n = df.loc[split_mask, 'weights'].sum()
            per_label_conf = df[split_mask].groupby('label')['weights'].sum()
            assert len(per_label_conf) == c, (
                f'{split} split only has {len(per_label_conf)}/{c} labels')
            scaling = (n / c) / per_label_conf[df.loc[split_mask, 'label']]
            df.loc[split_mask, 'weights'] *= scaling.to_numpy()
            w_sum = df.loc[split_mask, 'weights'].sum()
            assert np.isclose(w_sum, n), (
                f'Expected {split} weights to sum to {n}, got {w_sum} instead')

        # error checking
        assert (df['weights'] > 0).all()

    return df, label_names, split_to_locs


def recall_from_confusion_matrix(
        confusion_matrix: np.ndarray,
        label_names: Sequence[str],
        ) -> dict[str, float]:
    """
    Args:
        confusion_matrix: np.ndarray, shape [n_classes, n_classes], type int
            C[i, j] = # of samples with true label i, predicted as label j
        label_names: list of str, label names in order by label id

    Returns: dict, label_name => recall
    """
    
    result = {
        label_name: confusion_matrix[i, i] / (confusion_matrix[i].sum() + 1e-8)
        for i, label_name in enumerate(label_names)
    }
    return result



"""

visualize_detector_output.py

Render images with bounding boxes annotated on them to a folder, based on a
detector output result file (.json), optionally writing an HTML index file.

"""

#%% Imports

import argparse
import json
import os
import random
import sys
from multiprocessing.pool import ThreadPool
from multiprocessing.pool import Pool
from typing import List
from functools import partial

from tqdm import tqdm

from data_management.annotations.annotation_constants import detector_bbox_category_id_to_name
from md_visualization import visualization_utils as vis_utils
from md_utils.ct_utils import get_max_conf
from md_utils import write_html_image_list
from detection.run_detector import get_typical_confidence_threshold_from_results


#%% Constants

# This will only be used if a category mapping is not available in the results file.
DEFAULT_DETECTOR_LABEL_MAP = {
    str(k): v for k, v in detector_bbox_category_id_to_name.items()
}


#%% Support functions

def _render_image(entry,
                 detector_label_map,classification_label_map,
                 confidence_threshold,classification_confidence_threshold,
                 render_detections_only,preserve_path_structure,out_dir,images_dir,
                 output_image_width):
    """
    Internal function for rendering a single image.
    """
    
    rendering_result = {'failed_image':False,'missing_image':False,
                        'skipped_image':False,'annotated_image_path':None,
                        'max_conf':None,'file':entry['file']}
    
    image_id = entry['file']

    if 'failure' in entry and entry['failure'] is not None:
        rendering_result['failed_image'] = True
        return rendering_result

    assert 'detections' in entry and entry['detections'] is not None
    
    max_conf = get_max_conf(entry)
    rendering_result['max_conf'] = max_conf
    
    if (max_conf < confidence_threshold) and render_detections_only:
        rendering_result['skipped_image'] = True
        return rendering_result
    
    image_obj = os.path.join(images_dir, image_id)
    if not os.path.exists(image_obj):
        print(f'Image {image_id} not found in images_dir')
        rendering_result['missing_image'] = True
        return rendering_result

    # If output_image_width is -1 or None, this will just return the original image
    image = vis_utils.resize_image(
        vis_utils.open_image(image_obj), output_image_width)

    vis_utils.render_detection_bounding_boxes(
        entry['detections'], image, 
        label_map=detector_label_map,
        classification_label_map=classification_label_map,
        confidence_threshold=confidence_threshold,
        classification_confidence_threshold=classification_confidence_threshold)

    if not preserve_path_structure:
        for char in ['/', '\\', ':']:
            image_id = image_id.replace(char, '~')
        annotated_img_path = os.path.join(out_dir, f'anno_{image_id}')
    else:
        assert not os.path.isabs(image_id), "Can't preserve paths when operating on absolute paths"
        annotated_img_path = os.path.join(out_dir, image_id)
        os.makedirs(os.path.dirname(annotated_img_path),exist_ok=True)
        
    image.save(annotated_img_path)
    rendering_result['annotated_image_path'] = annotated_img_path        

    return rendering_result


#%% Main function

def visualize_detector_output(detector_output_path,
                              out_dir,
                              images_dir,
                              confidence_threshold=0.15,
                              sample=-1,
                              output_image_width=700,
                              random_seed=None,
                              render_detections_only=False,
                              classification_confidence_threshold=0.1,
                              html_output_file=None,
                              html_output_options=None,
                              preserve_path_structure=False,
                              parallelize_rendering=False,
                              parallelize_rendering_n_cores=10,
                              parallelize_rendering_with_threads=True):
    
    """
    Draws bounding boxes on images given the output of a detector.

    Args:
        detector_output_path (str): path to detector output .json file
        out_dir (str): path to directory for saving annotated images
        images_dir (str): folder where the images live; filenames in 
            [detector_output_path] should be relative to [image_dir]
        confidence_threshold (float, optional): threshold above which detections will be rendered
        sample (int, optional): maximum number of images to render, -1 for all
        output_image_width (int, optional): width in pixels to resize images for display,
            preserving aspect ration; set to -1 to use original image width
        random_seed (int, optional): seed to use for choosing images when sample != -1
        render_detections_only (bool): only render images with above-threshold detections
        classification_confidence_threshold (float, optional): only show classifications
            above this threshold; does not impact whether images are rendered, only whether
            classification labels (not detection categories) are displayed
        html_output_file (str, optional): output path for an HTML index file (not written
            if None)
        html_output_options (dict, optional): HTML formatting options; see write_html_image_list 
            for details
        preserve_path_structure (bool, optional): if False (default), writes images to unique
            names in a flat structure in the output folder; if True, preserves relative paths
            within the output folder
        parallelize_rendering (bool, optional): whether to use concurrent workers for rendering
        parallelize_rendering_n_cores (int, optional): number of concurrent workers to use 
            (ignored if parallelize_rendering is False)
        parallelize_rendering_with_threads (bool, optional): determines whether we use
            threads (True) or processes (False) for parallelization (ignored if parallelize_rendering 
            is False)

    Returns:
        list: list of paths to annotated images
    """
    
    assert os.path.exists(detector_output_path), \
        'Detector output file does not exist at {}'.format(detector_output_path)

    assert os.path.isdir(images_dir), \
        'Image folder {} is not available'.format(images_dir)

    os.makedirs(out_dir, exist_ok=True)


    ##%% Load detector output

    with open(detector_output_path) as f:
        detector_output = json.load(f)
    assert 'images' in detector_output, (
        'Detector output file should be a json with an "images" field.')
    images = detector_output['images']
    
    if confidence_threshold is None:
        confidence_threshold = get_typical_confidence_threshold_from_results(detector_output)
        
    assert confidence_threshold >= 0 and confidence_threshold <= 1, (
        f'Confidence threshold {confidence_threshold} is invalid, must be in (0, 1).')
    
    if 'detection_categories' in detector_output:
        print('Using custom label mapping')
        detector_label_map = detector_output['detection_categories']
    else:
        detector_label_map = DEFAULT_DETECTOR_LABEL_MAP

    num_images = len(images)
    print(f'Detector output file contains {num_images} entries.')

    if sample > 0:
        assert num_images >= sample, (
            f'Sample size {sample} greater than number of entries '
            f'({num_images}) in detector result.')

        if random_seed is not None:
            images = sorted(images, key=lambda x: x['file'])
            random.seed(random_seed)

        random.shuffle(images)
        images = sorted(images[:sample], key=lambda x: x['file'])
        print(f'Sampled {len(images)} entries from the detector output file.')


    ##%% Load images, annotate them and save

    print('Rendering detections above a confidence threshold of {}'.format(
        confidence_threshold))
    
    classification_label_map = None
    
    if 'classification_categories' in detector_output:
        classification_label_map = detector_output['classification_categories']
        
    rendering_results = []
    
    if parallelize_rendering:
        
        if parallelize_rendering_with_threads:
            worker_string = 'threads'
        else:
            worker_string = 'processes'
            
        if parallelize_rendering_n_cores is None:
            if parallelize_rendering_with_threads:
                pool = ThreadPool()
            else:
                pool = Pool()
        else:
            if parallelize_rendering_with_threads:
                pool = ThreadPool(parallelize_rendering_n_cores)
            else:
                pool = Pool(parallelize_rendering_n_cores)
            print('Rendering images with {} {}'.format(parallelize_rendering_n_cores,
                                                       worker_string))            
        rendering_results = list(tqdm(pool.imap(
                                 partial(_render_image,detector_label_map=detector_label_map,
                                         classification_label_map=classification_label_map,
                                         confidence_threshold=confidence_threshold,
                                         classification_confidence_threshold=classification_confidence_threshold,
                                         render_detections_only=render_detections_only,
                                         preserve_path_structure=preserve_path_structure,
                                         out_dir=out_dir,
                                         images_dir=images_dir,
                                         output_image_width=output_image_width),
                                 images), total=len(images)))
        
    else:
                
        for entry in tqdm(images):
            
            rendering_result = _render_image(entry,detector_label_map,classification_label_map,
                                            confidence_threshold,classification_confidence_threshold,
                                            render_detections_only,preserve_path_structure,out_dir,
                                            images_dir,output_image_width)
            rendering_results.append(rendering_result)
        
    # ...for each image
    
    failed_images = [r for r in rendering_results if r['failed_image']]
    missing_images = [r for r in rendering_results if r['missing_image']]
    skipped_images = [r for r in rendering_results if r['skipped_image']]
    
    print('Skipped {} failed images (of {})'.format(len(failed_images),len(images)))
    print('Skipped {} missing images (of {})'.format(len(missing_images),len(images)))
    print('Skipped {} below-threshold images (of {})'.format(len(skipped_images),len(images)))
    
    print(f'Rendered detection results to {out_dir}')

    annotated_image_paths = [r['annotated_image_path'] for r in rendering_results if \
                             r['annotated_image_path'] is not None]
    
    if html_output_file is not None:
        
        html_dir = os.path.dirname(html_output_file)
        
        html_image_info = []
        
        for r in rendering_results:
            d = {}
            annotated_image_path_relative = os.path.relpath(r['annotated_image_path'],html_dir)
            d['filename'] = annotated_image_path_relative
            d['textStyle'] = \
             'font-family:verdana,arial,calibri;font-size:80%;' + \
                 'text-align:left;margin-top:20;margin-bottom:5'
            d['title'] = '{} (max conf: {})'.format(r['file'],r['max_conf'])
            html_image_info.append(d)
            
        _ = write_html_image_list.write_html_image_list(html_output_file,html_image_info,
                                                    options=html_output_options)
        
    return annotated_image_paths


#%% Command-line driver

def main():
    
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Annotate the bounding boxes predicted by a detector above '
                    'some confidence threshold, and save the annotated images.')
    parser.add_argument(
        'detector_output_path', type=str,
        help='Path to json output file of the detector')
    parser.add_argument(
        'out_dir', type=str,
        help='Path to directory where the annotated images will be saved. '
             'The directory will be created if it does not exist.')
    parser.add_argument(
        '-c', '--confidence', type=float, default=0.15,
        help='Value between 0 and 1, indicating the confidence threshold '
             'above which to visualize bounding boxes')
    parser.add_argument(
        '-i', '--images_dir', type=str, default=None,
        help='Path to a local directory where images are stored. This '
             'serves as the root directory for image paths in '
             'detector_output_path.')
    parser.add_argument(
        '-n', '--sample', type=int, default=-1,
        help='Number of images to be annotated and rendered. Set to -1 '
             '(default) to annotate all images in the detector output file. '
             'There may be fewer images if some are not found in images_dir.')
    parser.add_argument(
        '-w', '--output_image_width', type=int, default=700,
        help='Integer, desired width in pixels of the output annotated images. '
             'Use -1 to not resize. Default: 700.')
    parser.add_argument(
        '-r', '--random_seed', type=int, default=None,
        help='Integer, for deterministic order of image sampling')
    parser.add_argument(
        '-html', '--html_output_file', type=str, default=None,
        help='Filename to which we should write an HTML image index (off by default)')
    parser.add_argument(
        '--open_html_output_file', action='store_true',
        help='Open the .html output file when done')
    parser.add_argument(
        '-do', '--detections_only', action='store_true',
        help='Only render images with above-threshold detections (by default, '
             'both empty and non-empty images are rendered).')
    parser.add_argument(
        '-pps', '--preserve_path_structure', action='store_true',
        help='Preserve relative image paths (otherwise flattens and assigns unique file names)')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()
    visualize_detector_output(
        detector_output_path=args.detector_output_path,
        out_dir=args.out_dir,
        confidence_threshold=args.confidence,
        images_dir=args.images_dir,
        sample=args.sample,
        output_image_width=args.output_image_width,
        random_seed=args.random_seed,
        render_detections_only=args.detections_only,
        preserve_path_structure=args.preserve_path_structure,
        html_output_file=args.html_output_file)

    if args.html_output_file is not None and args.open_html_output_file:
        from md_utils.path_utils import open_file
        open_file(args.html_output_file)

if __name__ == '__main__':
    main()


#%% Interactive driver

if False:
    
    pass

    #%%
    
    detector_output_path = os.path.expanduser('~/postprocessing/bellevue-camera-traps/bellevue-camera-traps-2023-12-05-v5a.0.0/combined_api_outputs/bellevue-camera-traps-2023-12-05-v5a.0.0_detections.json')
    out_dir = r'g:\temp\preview'
    images_dir = r'g:\camera_traps\camera_trap_images'
    confidence_threshold = 0.15
    sample = 50
    output_image_width = 700
    random_seed = 1
    render_detections_only = True
    classification_confidence_threshold = 0.1
    html_output_file = os.path.join(out_dir,'index.html')
    html_output_options = None
    preserve_path_structure = False
    parallelize_rendering = True
    parallelize_rendering_n_cores = 10
    parallelize_rendering_with_threads = False

    _ = visualize_detector_output(detector_output_path,
                              out_dir,
                              images_dir,
                              confidence_threshold,
                              sample,
                              output_image_width,
                              random_seed,
                              render_detections_only,
                              classification_confidence_threshold,
                              html_output_file,
                              html_output_options,
                              preserve_path_structure,
                              parallelize_rendering,
                              parallelize_rendering_n_cores,
                              parallelize_rendering_with_threads)
    
    from md_utils.path_utils import open_file
    open_file(html_output_file)


# 
# api_frontend.py
#
# Defines the Flask app, which takes requests (one or more images) from
# remote callers and pushes the images onto the shared Redis queue, to be processed
# by the main service in api_backend.py .
#

#%% Imports

import os
import json
import time
import uuid
import redis
import shutil
import argparse
import traceback

from io import BytesIO
from flask import Flask, Response, jsonify, make_response, request
from requests_toolbelt.multipart.encoder import MultipartEncoder

import md_visualization.visualization_utils as vis_utils
import config


#%% Initialization

app = Flask(__name__)
db = redis.StrictRedis(host=config.REDIS_HOST, port=config.REDIS_PORT)


#%% Support functions

def _make_error_object(error_code, error_message):
    
    # Make a dict that the request_processing_function can return to the endpoint 
    # function to notify it of an error
    return {
        'error_message': error_message,
        'error_code': error_code
    }


def _make_error_response(error_code, error_message):

    return make_response(jsonify({'error': error_message}), error_code)


def has_access(request):
    
    if not os.path.exists(config.API_KEYS_FILE):
        return True
    else:
        if not request.headers.get('key'):
            print('Key header not available')
            return False
        else:
            API_key = request.headers.get('key').strip().lower()
            with open(config.API_KEYS_FILE, "r") as f:
                for line in f:
                    valid_key = line.strip().lower()
                    if valid_key == API_key:
                        return True

    return False


def check_posted_data(request):
 
    files = request.files
    params = request.args

    # Verify that the content uploaded is not too big
    #
    # request.content_length is the length of the total payload
    content_length = request.content_length
    if not content_length:
        return _make_error_object(411, 'No image(s) were sent, or content length cannot be determined.')
    if content_length > config.MAX_CONTENT_LENGTH_IN_MB * 1024 * 1024:
        return _make_error_object(413, ('Payload size {:.2f} MB exceeds the maximum allowed of {} MB. '
                    'Please upload fewer or more compressed images.').format(
            content_length / (1024 * 1024), config.MAX_CONTENT_LENGTH_IN_MB))

    render_boxes = True if params.get('render', '').lower() == 'true' else False

    if 'min_confidence' in params:
        return_confidence_threshold = float(params['min_confidence'])
        print('runserver, post_detect_sync, user specified detection confidence: ', return_confidence_threshold)
        if return_confidence_threshold < 0.0 or return_confidence_threshold > 1.0:
            return _make_error_object(400, 'Detection confidence threshold {} is invalid, should be between 0.0 and 1.0.'.format(
                return_confidence_threshold))
    else:
        return_confidence_threshold = config.DEFAULT_CONFIDENCE_THRESHOLD
        
    if 'min_rendering_confidence' in params:
        rendering_confidence_threshold = float(params['min_rendering_confidence'])
        print('runserver, post_detect_sync, user specified rendering confidence: ', rendering_confidence_threshold) 
        if rendering_confidence_threshold < 0.0 or rendering_confidence_threshold > 1.0:
            return _make_error_object(400, 'Rendering confidence threshold {} is invalid, should be between 0.0 and 1.0.'.format(
                rendering_confidence_threshold))
    else:
        rendering_confidence_threshold =  config.DEFAULT_RENDERING_CONFIDENCE_THRESHOLD
    
    # Verify that the number of images is acceptable
    num_images = sum([1 if file.content_type in config.IMAGE_CONTENT_TYPES else 0 for file in files.values()])
    print('runserver, post_detect_sync, number of images received: ', num_images)

    if num_images > config.MAX_IMAGES_ACCEPTED:
        return _make_error_object(413, 'Too many images. Maximum number of images that can be processed in one call is {}.'.format(config.MAX_IMAGES_ACCEPTED))
    elif num_images == 0:
        return _make_error_object(400, 'No image(s) of accepted types (image/jpeg, image/png, application/octet-stream) received.')
    
    return {
        'render_boxes': render_boxes,
        'return_confidence_threshold': return_confidence_threshold,
        'rendering_confidence_threshold': rendering_confidence_threshold
    }

# ...def check_posted_data(request)
    

#%% Main loop

@app.route(config.API_PREFIX + '/detect', methods = ['POST'])
def detect_sync():
    
    if not has_access(request):
        print('Access denied, please provide a valid API key')
        return _make_error_response(403, 'Access denied, please provide a valid API key')

    # Check whether the request_processing_function had an error
    post_data = check_posted_data(request)
    if post_data.get('error_code', None) is not None:
        return _make_error_response(post_data.get('error_code'), post_data.get('error_message'))

    render_boxes = post_data.get('render_boxes')
    return_confidence_threshold = post_data.get('return_confidence_threshold')
    rendering_confidence_threshold = post_data.get('rendering_confidence_threshold')
  
    redis_id = str(uuid.uuid4())
    d = {'id': redis_id, 'render_boxes': render_boxes, 'return_confidence_threshold': return_confidence_threshold}
    temp_direc = os.path.join(config.TEMP_FOLDER, redis_id)
    
    try:
        
        try:
            # Write images to temporary files
            #
            # TODO: read from memory rather than using intermediate files
            os.makedirs(temp_direc,exist_ok=True)
            for name, file in request.files.items():
                if file.content_type in config.IMAGE_CONTENT_TYPES:
                    filename = request.files[name].filename
                    image_path = os.path.join(temp_direc, filename)
                    print('Saving image {} to {}'.format(name,image_path))
                    file.save(image_path)
                    assert os.path.isfile(image_path),'Error creating file {}'.format(image_path)
        
        except Exception as e:
            return _make_error_object(500, 'Error saving images: ' + str(e))
        
        # Submit the image(s) for processing by api_backend.py, who is waiting on this queue
        db.rpush(config.REDIS_QUEUE_NAME, json.dumps(d))
        
        while True:
            
            # TODO: convert to a blocking read and eliminate the sleep() statement in this loop
            result = db.get(redis_id)
            
            if result:
                
                result = json.loads(result.decode())
                print('Processing result {}'.format(str(result)))
                
                if result['status'] == 200:
                    detections = result['detections']
                    db.delete(redis_id)
                
                else:
                    db.delete(redis_id)
                    print('Detection error: ' + str(result))
                    return _make_error_response(500, 'Detection error: ' + str(result))

                try:
                    print('detect_sync: postprocessing and sending images back...')
                    fields = {
                        'detection_result': ('detection_result', json.dumps(detections), 'application/json'),
                    }

                    if render_boxes and result['status'] == 200:

                        print('Rendering images')

                        for image_name, detections in detections.items():
                            
                            #image = Image.open(os.path.join(temp_direc, image_name))
                            image = open(f'{temp_direc}/{image_name}', "rb")
                            image = vis_utils.load_image(image)
                            width, height = image.size

                            _detections = []
                            for d in detections:
                                y1,x1,y2,x2 = d[0:4]
                                width = x2 - x1
                                height = y2 - y1
                                bbox = [x1,y1,width,height]
                                _detections.append({'bbox': bbox, 'conf': d[4], 'category': d[5]}) 
                            
                            vis_utils.render_detection_bounding_boxes(_detections, image, 
                            confidence_threshold=rendering_confidence_threshold)
                            
                            output_img_stream = BytesIO()
                            image.save(output_img_stream, format='jpeg')
                            output_img_stream.seek(0)
                            fields[image_name] = (image_name, output_img_stream, 'image/jpeg')
                        print('Done rendering images')
                        
                    m = MultipartEncoder(fields=fields)                    
                    return Response(m.to_string(), mimetype=m.content_type)

                except Exception as e:
                    
                    print(traceback.format_exc())
                    print('Error returning result or rendering the detection boxes: ' + str(e))

                finally:
                    
                    try:
                        print('Removing temporary files')
                        shutil.rmtree(temp_direc)
                    except Exception as e:
                        print('Error removing temporary folder {}: {}'.format(temp_direc,str(e)))
                    
            else:
                time.sleep(0.005)
                
            # ...if we do/don't have a request available on the queue
            
        # ...while(True)

    except Exception as e:
        
        print(traceback.format_exc())
        return _make_error_object(500, 'Error processing images: ' + str(e))

# ...def detect_sync()


#%% Command-line driver
    
if __name__ == '__main__':
    
    parser = argparse.ArgumentParser(description='api frontend')

    # use --non-docker if you are testing without Docker
    #
    # python api_frontend.py --non-docker
    parser.add_argument('--non-docker', action="store_true", default=False)
    args = parser.parse_args()

    if args.non_docker:
        app.run(host='0.0.0.0', port=5050)
    else:
        app.run()
